"""
File containing the Sandbox class.

TODO: Handle sys.argv

"""

import sys
import io
from itertools import zip_longest
from unittest.mock import patch

from pedal.core.feedback_category import FeedbackCategory
from pedal.core.report import MAIN_REPORT
from pedal.utilities.exceptions import ExpandedTraceback, improve_builtin_exceptions
from pedal.sandbox.data import SandboxVariable, SandboxContextKind, \
    SandboxContext, SandboxModules
from pedal.sandbox import mocked
from pedal.sandbox.constants import TOOL_NAME
from pedal.sandbox.feedbacks import runtime_error, EXCEPTION_FF_MAP
from pedal.sandbox.exceptions import SandboxHasNoFunction, SandboxHasNoVariable
from pedal.sandbox.timeout import timeout
from pedal.sandbox.result import SandboxResult
from pedal.sandbox.tracer import TRACER_STYLES


class Sandbox:
    """
    Args:
        report (:py:class:`~pedal.core.report.Report`): The report that this
            Sandbox is attached to.

    Attributes:
        data (dict[str, Any]): The namespace that the context are occurring in.
            Note that this is mutable.
        result: TODO
        output (list[str]): The list of strings that have been printed to the
            console by :py:func:`print`. Note that line endings have been removed
            using :py:func:`string.rstrip`.
        inputs (list[str]): The list of strings that will be passed to
            :py:func:`input`.
        raw_output (str): A concatenated string of all the text that was printed
            to console. No changes have been made to the text.
        exception (Exception or None): The exception that occurred during the
            most recent execution, or None if nothing happened.
        feedback: TODO
        modules.plots: TODO
        modules.turtles: TODO
        target: TODO
        allowed_time (int): How long to allow before stopping execution.
        tracer_style (str): TODO
        _context (list[SandboxContext]): The history of executions made in
            this sandbox.
        _next_context_id (int): The ID of the next execution context.
        _current_patches (list[Patch]): A stack of patches that are currently
            being used. This allows recursive patching behavior.
        MAXIMUM_TEMPORARY_LENGTH (int): How long to allow arguments to be before
            turning them into temporary variables.
    """

    MAXIMUM_TEMPORARY_LENGTH = 25

    def __init__(self, report=MAIN_REPORT):
        self.report = report
        # Namespace
        self.data = {}
        self.result = None
        self.exception = None
        self.feedback = None
        # Contextualization
        self._context = []
        self._current_context = None
        self._next_context_id = 0
        self._context_group_start = []
        # Patching
        self._current_patches = []
        self._current_stdout = []
        # Temporary Variables
        self._temporary_variables = set()
        self._backup_variables = {}
        # Modules
        self._module_overrides = {}
        self.modules = SandboxModules()
        self.clear_mocks()
        self.clear_data()
        # Inputs
        self.inputs = []
        # Outputs
        self.raw_output = ""
        self.output = []
        # Target
        self.target = None

        # Proxying results
        self.result_proxy_class = SandboxResult
        # Show the full traceback
        self.full_traceback = False
        # Use threading?
        self.threaded = False
        self.allowed_time = 3
        # Tracer Styles
        self.tracer_style = 'none'

    ############################################################################
    # Execution (run/call/eval)

    def _execute_with_timeout(self, code, filename, kind):
        """
        Execute the given code, but stop after `self.allowed_time`.
        Args:
            code (str):
            filename (str):
            kind (:py:class:`pedal.sandbox.sandbox_mixins.SandboxContextKind`):

        Returns:
            :py:class:`pedal.sandbox.sandbox.Sandbox`
        """
        try:
            return timeout(self.allowed_time, self._execute,
                           code, filename, kind, False)
        except TimeoutError as timeout_exception:
            self._stop_patches()
            self._capture_exception(timeout_exception, sys.exc_info(),
                                    code, filename)
            return self

    def _execute(self, code, filename, kind, threaded, **meta):
        # Handle any threading if necessary
        if threaded:
            return self._execute_with_timeout(code, filename, kind)

        self.clear_exception()
        context = SandboxContext(self._next_context_id, code, filename, kind,
                                 self.target, [], "",
                                 self.exception, self.report.submission, **meta)
        self._context.append(context)

        # Patch in dangerous built-ins
        # Override builtins and mock stuff out
        self._start_mocking(context)
        try:
            # TODO: Support CaitNode and Ast (needs skulpt to support compile better)
            compiled_code = compile(code, filename, 'exec')
            with self.trace.as_filename(filename, code):
                exec(compiled_code, self.data)
        except Exception as user_exception:
            self._stop_mocking(context)
            self._capture_exception(user_exception, sys.exc_info(),
                                    code, filename)
        else:
            self._stop_mocking(context)

        self._next_context_id += 1
        return self

    def run(self, code=None, filename=None, inputs=None, threaded=None,
            after=None, before=None):
        """
        If both ``code`` and ``filename`` are None, then the submission's
        main file will be executed. If ``code`` is given but ``filename`` is
        not, then it is assumed to be instructor code.

        Args:
            code (str or :py:class:`~pedal.cait.cait_node.CaitNode` or None):
                The code to execute.
            filename (str or None): The filename to use for this code.
            inputs (list[str]): Optional inputs to be passed to
                :py:func:`~pedal.sandbox.Sandbox.set_input`.
            threaded (bool): Whether or not to run this code in a threaded
                environment, which allows for timeouts.
            after (str): Code to run after this code (without a filename).
            before (str): Code to run before this code (without a filename).


        Returns:
            Sandbox: This sandbox instance.
        """
        if threaded is None:
            threaded = self.threaded
        if code is None:
            if filename is None:
                filename = self.report.submission.main_file
            code = self.report.submission.files[filename]
        elif filename is None:
            filename = self.report.submission.instructor_file
        if inputs is not None:
            self.set_input(inputs)
        if before is not None:
            self._execute(before, filename, SandboxContextKind.RUN, threaded)
        self.target = None
        self._execute(code, filename, SandboxContextKind.RUN, threaded)
        if after is not None:
            self._execute(after, filename, SandboxContextKind.RUN, threaded)
        return self

    def call(self, function, *args, target="_", threaded=None,
             inputs=None, function_kwargs=None,
             args_locals=None, kwargs_locals=None, **kwargs):
        """
        Execute the given function from the student's namespace.

        The ``args_locals`` and ``kwargs_locals`` values allow you to
        use student's local variables as arguments, instead of literal values.
        They actually support any arbitrary Python code, it will be injected
        without modification.

        Args:
            function (str): The name of the function to call.
            *args (Any): Any number of positional arguments to be passed to
                the called function.
            target (str): The variable to assign the result of this call to.
                Defaults to ``"_"``.
            threaded (bool): Whether or not to run this code in a threaded
                environment, which allows for timeouts.
            inputs (str or list[str]):
            function_kwargs (dict[str, Any]): Additional keyword arguments
                that could not be passed in directly via ``**kwargs`` (perhaps
                because they conflict with an existing parameter like
                ``target``).
            args_locals (list[str]): A list of names (or any valid Python
                expression) that will be passed as positional arguments to
                the function (as opposed to actual values). If any value is
                None (or if the list is too short), then the corresponding
                position argument from ``*args`` will be used instead.
            kwargs_locals (dict[str, str]): A dictionary of keyword argument
                names mapped to local names (or any valid Python expression),
                that will be used as keyword parameters similar to
                ``args_locals``.
            **kwargs (): Additional keyword arguments passed to the called
                function.

        Returns:
            Exception or :py:class:`~pedal.sandbox.sandbox.SandboxResult`: The
                result of calling the function will be returned, proxied behind
                a SandboxResult (which attempts to perfectly emulate that
                value). If the function call failed, the exception will be
                returned instead.
        """
        # Confirm that the function_name exists
        if function not in self.functions:
            if function not in self.data:
                self.exception = SandboxHasNoVariable(
                    "The function {function} does not exist.".format(function=function)
                )
            else:
                self.exception = SandboxHasNoFunction(
                    "The variable {function} is not a function.".format(function=function)
                )
            return self.exception
        # Handle convenience setup
        if inputs is not None:
            self.set_input(inputs)
        if function_kwargs is not None:
            kwargs.update(function_kwargs)
        if args_locals is None:
            args_locals = []
        if kwargs_locals is None:
            kwargs_locals = []
        # Make the call and evaluate it
        self.target = target
        actual, student, arguments = self._construct_call(function, args, kwargs,
                                                          args_locals, kwargs_locals,
                                                          target)
        context_id = self._next_context_id
        self._execute(actual, self.report.submission.instructor_file,
                      SandboxContextKind.CALL, threaded=threaded,
                      called=function, args=arguments)
        self._purge_temporaries()
        return self._handle_result(target, context_id)

    def evaluate(self, code, target="_", threaded=None):
        """
        Evaluates the given code and assigns the result to the given target.
        Will cause an error if ``code`` is not a valid expression.

        # TODO: Clean up syntax error.

        Args:
            code (str or :py:class:`~pedal.cait.cait_node.CaitNode`):
                The code to execute. If a CaitNode, then that will be executed
                directly instead of being compiled.
            target (str): The name of the variable to assign the result to.
                Note that the result is also returned by this function.
            threaded (bool): Whether or not to run this code in a threaded
                environment, which allows for timeouts.

        Returns:
            Exception or :py:class:`~pedal.sandbox.sandbox.SandboxResult`: The
                result of evaluating the code will be returned, proxied behind
                a SandboxResult (which attempts to perfectly emulate that
                value). If the function call failed, the exception will be
                returned instead.
        """
        self.target = target
        code = f"{target} = {code}"
        context_id = self._next_context_id
        self._execute(code, self.report.submission.instructor_file,
                      SandboxContextKind.EVAL, threaded=threaded)
        return self._handle_result(target, context_id)

    def _handle_result(self, target, context_id):
        """ Determine the appropriate return value (either an exception or the
        value stored in target. Also updates self.result. """
        if self.exception is None:
            self.result = self.data[target]
            if self.result_proxy_class is not None:
                self.result = self.result_proxy_class(self.result,
                                                      context_id=context_id,
                                                      sandbox=self)
            return self.result
        else:
            if self.result_proxy_class is not None:
                self.exception = self.result_proxy_class(self.exception,
                                                         context_id=context_id,
                                                         sandbox=self)
            return self.exception

    ############################################################################
    # Context (history of executions)

    def get_context(self, context_id: int = None):
        """
        Retrieve the most recent contexts.

        Returns:
            list[:py:class:`pedal.sandbox.sandbox_mixins.SandboxContext`]: The
                most recent execution, or executions if currently in a group.

        """
        # TODO: Test off-by-one-errors
        if context_id is None:
            if not self._context_group_start:
                return self._context[-1:]
            else:
                return self._context[self._context_group_start[-1]:]
        else:
            if not self._context_group_start:
                return [self._context[context_id]]
            else:
                for past_context_group_starts in self._context_group_start[::-1]:
                    if past_context_group_starts < context_id+1:
                        return self._context[past_context_group_starts:context_id+1]
                return self._context[:context_id+1]

    def _guess_context(self, target_name):
        """
        Tries to figure out the best context for the given target_name, by
        first checking whether the target was ever used in a CALL or EVAL.
        If it fails to find it, then it uses the most recent RUN. Otherwise,
        it just just returns None.

        Returns:
            :py:class:`pedal.sandbox.data.SandboxContext` or None: The found
                context, or None.
        """
        # Was it a target for a CALL or EVAL?
        for context in self._context[::-1]:
            if context.kind in (SandboxContextKind.EVAL, SandboxContextKind.CALL):
                if context.target == target_name:
                    return context
        # Just get the last run
        for context in self._context[::-1]:
            if context.kind == SandboxContextKind.RUN:
                return context
        # Okay just give up
        return None

    def start_grouping_context(self):
        """ Any subsequent executions will be grouped. """
        self._context_group_start.append(self._next_context_id)

    def stop_grouping_context(self):
        """ Stop grouping subsequent executions. """
        self._context_group_start.pop()

    def clear_context(self):
        """ Removes the history of any previous executions. """
        self._context_group_start.clear()
        self._context.clear()

    ############################################################################
    # Tracing
    def _set_tracer_style(self, tracer_style):
        if tracer_style is None:
            tracer_style = 'none'
        self._tracer_style = tracer_style.lower()
        self.trace = TRACER_STYLES[tracer_style.lower()]()

    def _get_tracer_style(self):
        return self._tracer_style

    tracer_style = property(_get_tracer_style, _set_tracer_style)

    def clear_tracer(self):
        """ Stop tracing in this sandbox. """
        self.tracer_style = 'none'

    ############################################################################
    # Exception Handling

    def _capture_exception(self, exception, exc_info, code, filename):
        self.exception = improve_builtin_exceptions(exception)
        # Load in any extra data
        context = self.get_context()
        line_offsets = {}
        show_filenames = {filename} - {self.report.submission.instructor_file}
        hide_filenames = {filename}
        files = self.report.submission.get_files_lines()
        if filename not in files:
            files[filename] = code.split("\n")
        if self.report.submission is not None:
            lines = self.report.submission.get_lines()
            show_filenames.update(self.report.submission.files.keys())
            line_offsets = self.report.submission.line_offsets
        else:
            lines = code.split("\n")
        # Create a better traceback
        traceback = ExpandedTraceback(self.exception, exc_info,
                                      self.full_traceback,
                                      hide_filenames,
                                      line_offsets,
                                      show_filenames,
                                      lines, files)
        if filename == self.report.submission.instructor_file:
            priority = FeedbackCategory.SPECIFICATION
        else:
            priority = FeedbackCategory.RUNTIME

        runtime_error_function = EXCEPTION_FF_MAP.get(type(self.exception),
                                                      runtime_error)
        self.feedback = runtime_error_function(exception=self.exception, context=[context],
                                               traceback=traceback, location=traceback.line_number,
                                               report=self.report, priority=priority)
        return False

    def clear_exception(self):
        """ Removes the latest exception information """
        self.exception = None
        self.feedback = None

    ############################################################################
    # Patching and Mocking

    def _start_mocking(self, context: SandboxContext):
        """ Mock input, output, builtins, and modules """
        # Handle input tracking
        self.mock_function('input', self._track_inputs(context.inputs))
        # Override builtin functions
        self.reset_builtins()
        builtins = self._module_overrides.pop('__builtins__', {})
        for name, value in builtins.items():
            if value is True:
                self.data['__builtins__'][name] = mocked.ORIGINAL_BUILTINS[name]
                self.data[name] = mocked.ORIGINAL_BUILTINS[name]
            elif value is False:
                self.data['__builtins__'][name] = mocked.disabled_builtin(name)
                self.data[name] = mocked.disabled_builtin(name)
            else:
                self.data['__builtins__'][name] = value
                self.data[name] = value
        # Override sys modules
        overridden_modules = sys.modules.copy()
        for name, value in self._module_overrides.items():
            if value is not True:
                overridden_modules[name] = value
        self._module_overrides['__builtins__'] = builtins
        # And do the patches
        self._current_stdout.append(io.StringIO())
        self._start_patches(
            patch.dict('sys.modules', overridden_modules),
            patch('sys.stdout', self._current_stdout[-1]),
            patch('time.sleep', return_value=None),
        )

    def _stop_mocking(self, context: SandboxContext):
        """ Turn off any patches, store output """
        self._stop_patches()
        current_stdout = self._current_stdout.pop()
        self.append_output(current_stdout.getvalue(), context)

    # Patching Functionality
    def _start_patches(self, *patches):
        """ Helper function to start and keep track of multiple patches """
        self._current_patches.append(patches)
        for a_patch in patches:
            a_patch.start()

    def _stop_patches(self):
        """ Helper function to end any tracked patches """
        if not self._current_patches:
            return
        patches = self._current_patches.pop()
        for a_patch in patches:
            a_patch.stop()

    def clear_mocks(self):
        """ Removes any module or builtin overrides currently in effect. """
        self._module_overrides.clear()
        self.modules.clear()
        self.reset_default_overrides()

    def reset_default_overrides(self):
        """
        Resets the list of blocked functions and modules to its starting
        state. This blocks ``compile``, ``eval``, ``exec``, ``globals``,
        and provides mocked versions of ``open`` and ``import`` that are
        more restricted. It also blocks the ``pedal`` module.
        """
        self._module_overrides['__builtins__'] = {}
        self.block_function('compile')
        self.block_function('eval')
        self.block_function('exec')
        self.block_function('globals')
        self.mock_function('open', mocked._restricted_open)
        self.mock_function('__import__', mocked._restricted_import)
        # TODO: This breaks coverage for some reason; it's a builtin?
        self.block_module('pedal')
        self.mock_module('turtle', mocked.MockTurtle(), 'turtles')
        self.mock_module('matplotlib.pyplot', mocked.MockPlt(), 'plotting')

    def mock_function(self, function_name, new_version):
        self._module_overrides['__builtins__'][function_name] = new_version

    def allow_function(self, function_name):
        self._module_overrides['__builtins__'][function_name] = True

    def block_function(self, function_name):
        self._module_overrides['__builtins__'][function_name] = False

    def allow_module(self, module_name):
        """

        Args:
            module_name (str):

        Returns:

        """
        self._module_overrides[module_name] = True
        return self

    def mock_module(self, module_name, new_version, friendly_name=None):
        """ Create a mocked version of this module. """
        if friendly_name is None:
            friendly_name = module_name
        mocked_modules = self.modules.new_module(new_version, module_name, friendly_name)
        for name, mocked_version in mocked_modules.items():
            if name not in self._module_overrides or not self._module_overrides[name]:
                self._module_overrides[name] = mocked_version
        return self

    def block_module(self, module_name):
        """
        Prevent the module from being loaded in student code.

        Args:
            module_name (str): The name of the module to prevent, as it would
                be used by import (e.g., ``"matplotlib.pyplot"``).

        Returns:
            Sandbox: This sandbox.
        """
        self._module_overrides[module_name] = mocked.BlockedModule(module_name)
        return self

    ############################################################################
    # Code generation

    def _construct_call(self, function, args, kwargs, args_locals, kwargs_locals,
                        target):
        """ Turn the given strings into an actual function call string. """
        str_args = [arg_name if arg_name is not None else
                    self._make_temporary('arg', str(index), arg_value)
                    for index, (arg_value, arg_name)
                    in enumerate(zip_longest(args, args_locals))]
        str_kwargs = ["{}={}".format(key,
                                     self._make_temporary('kwarg', key, value))
                      if key not in kwargs_locals else
                      kwargs_locals[key]
                      for key, value in kwargs.items()]
        arguments = ", ".join(str_args + str_kwargs)
        call = f"{function}({arguments})"
        if target is None:
            actual = call
        else:
            actual = f"{target} = {call}"
        student_call = call if target == "_" else actual
        return actual, student_call, arguments

    def _purge_temporaries(self):
        """
        Delete any variables in the namespace that have been made as
        temporaries. This happens automatically after you execute code.
        """
        for key in self._temporary_variables:
            if key in self._backup_variables:
                self.data[key] = self._backup_variables[key]
            else:
                del self.data[key]
        self._temporary_variables.clear()

    def _make_temporary(self, category, name, value):
        """
        Create a temporary variable in the namespace for the given
        category/name. This is used to load arguments into the namespace to
        be used in function calls. Temporaries are only created if the value's
        repr length is too long, as defined by _is_long_value.

        Args:
            category (str): A categorical division for the temporary variable
                that can help keep the namespace distinctive - there are a
                few different kinds of categories (e.g., for regular positional
                args, star args, kwargs).
            name (str): A distinctive ID for this variable. The final variable
                name will be "_temporary_<category>_<name>".
            value: The value for this argument.
        Returns:
            str: The new name for the temporary variable.
        """
        if isinstance(value, SandboxVariable):
            return value.name
        if len(repr(value)) <= self.MAXIMUM_TEMPORARY_LENGTH:
            return repr(value)
        key = '_temporary_{}_{}'.format(category, name)
        if key in self.data:
            self._backup_variables[key] = self.data[key]
        self._temporary_variables.add(key)
        self.data[key] = value
        return key

    def make_safe_variable(self, name):
        """
        Tries to construct a safe variable name in the current namespace, based
        off the given one. This is accomplished by appending a "_" and a number
        of increasing value until no comparable name exists in the namespace.
        This is particularly useful when you want to create a variable name to
        assign to, but you are concerned that the user might have a variable
        with that name already, which their code relies on.

        Args:
            name (str): A desired target name.
        Returns:
            str: A safe target name, based off the given one.
        """
        current_addition = ""
        attempt_index = 2
        while name + current_addition in self.data:
            current_addition = "_{}".format(attempt_index)
            attempt_index += 1
        return name + current_addition


    ############################################################################
    # Data Namespace Management

    def clear_data(self):
        # Temporary data
        self.data.clear()
        self._temporary_variables.clear()
        self._backup_variables.clear()
        self.reset_builtins()

    def reset_builtins(self):
        self.data['__builtins__'] = {}
        for name, value in mocked._default_builtins.items():
            self.data['__builtins__'][name] = value

    def set_student_data(self, new_data):
        """
        Overwrites the existing student data with the keys and values from
        the ``new_data``. This copies the data over, but does not modify the
        reference to ``data``'s dictionary.

        Args:
            new_data (dict[str, Any]): The new data.

        Returns:

        """
        self.clear_data()
        for key, value in new_data.items():
            self.data[key] = value

    def get_names_by_type(self, type, exclude_builtins=True):
        """

        Args:
            type:
            exclude_builtins:

        Returns:

        """
        result = []
        for name, value in self.data.items():
            if isinstance(value, type):
                if exclude_builtins and name.startswith('__'):
                    continue
                result.append(name)
        return result

    def get_values_by_type(self, type, exclude_builtins=True):
        """

        Args:
            type:
            exclude_builtins:

        Returns:

        """
        names = self.get_names_by_type(type, exclude_builtins)
        return [self.data[name] for name in names]

    def get_variables_by_type(self, type, exclude_builtins=True):
        """

        Args:
            type:
            exclude_builtins:

        Returns:

        """
        names = self.get_names_by_type(type, exclude_builtins)
        return [(name, self.data[name]) for name in names]

    @property
    def functions(self):
        """
        Retrieve a list of all the callable names in the students' namespace.
        In other words, get a list of all the functions the student defined.

        Returns:
            list of callables
        """
        return {k: v for k, v in self.data.items() if callable(v)}

    @property
    def var(self):
        """

        Returns:

        """
        return {k: SandboxVariable(k, v) for k, v in self.data.items()}

    def __getitem__(self, item):
        value, exception = None, None
        try:
            value = self.data[item]
        except KeyError:
            exception = NameError(f"NameError: name '{item}' is not defined")
        filename = self.report.submission.instructor_file
        context_id = self._next_context_id
        best_context = self._guess_context(item)
        if best_context is None:
            context = SandboxContext(context_id, str(item), filename,
                                     SandboxContextKind.GETITEM, item, [], "",
                                     exception, self.report.submission)
        else:
            context = best_context.clone(context_id)
            context.target = str(item)
            context.exception = exception
        self._context.append(context)
        self._next_context_id += 1

        if self.result_proxy_class is not None:
            return self.result_proxy_class(value, context_id=context_id,
                                           sandbox=self)
        return value

    ############################################################################
    # Useful Dunders

    #def __repr__(self):
    #    return "Sandbox({})".format()

    def __str__(self):
        return "Sandbox(contexts={context_count})".format(context_count=len(self._context))

    ############################################################################
    # Input Handling

    def clear_input(self):
        """ Removes any inputs queued. """
        self.set_input(None)
        return self

    def set_input(self, inputs, clear=True):
        """
        Queues the given value as the next arguments to the `input` function.
        """
        if inputs is None:
            self.inputs = []
        if clear:
            self.inputs.clear()
        if isinstance(inputs, str):
            self.inputs.append(inputs)
        elif isinstance(inputs, (list, tuple)):
            self.inputs.extend(inputs)
        elif inputs is not None:
            self.inputs = inputs
        return self

    def _track_inputs(self, context_inputs):
        """
        Wraps an input function with a tracker.
        """

        def _input_tracker(*args, **kwargs):
            if args:
                prompt = args[0]
            else:
                prompt = ""
            print(prompt)
            if self.inputs:
                value_entered = self.inputs.pop(0)
            else:
                # TODO: Make this smarter, more elegant in choosing IF we should repeat 0
                value_entered = '0'
            self._context[-1].inputs.append(value_entered)
            #context_inputs.append(value_entered)
            return value_entered

        return _input_tracker

    ############################################################################
    # Output Handling

    def clear_output(self):
        """ Remove any output currently attached to this sandbox. """
        # Update outputs
        self.raw_output = ""
        self.output.clear()
        return self

    def append_output(self, raw_output, context):
        """
        Adds the string of `raw_output` to the current `raw_output` attribute.
        The added string will be split on newlines and rstripped to append
        to the `output` attribute.

        Args:
            raw_output (str): The new raw_output for the sandbox. To compute
                the `output` attribute, the system splits and rstrips at
                newlines.
            context (:py:class:`pedal.sandbox.data.SandboxContext`): The
                currently executing context, where this output will be recorded.
        """
        self.raw_output += raw_output
        context.output = raw_output
        if self.raw_output:
            lines = raw_output.rstrip().split("\n")
            lines = [line.rstrip() for line in lines]
            self.output.extend(lines)

    def clear(self):
        """ Removes any existing data in this sandbox. """
        self.clear_data()
        self.clear_context()
        self.clear_mocks()
        self.clear_exception()
        self.clear_input()
        self.clear_output()
        self.clear_tracer()



"""
Helper functions for traversing through the code structure.
"""

from pedal.cait import parse_program
from pedal.core.report import MAIN_REPORT
from pedal.utilities.operators import COMPARE_OP_NAMES, BOOL_OP_NAMES, BIN_OP_NAMES, UNARY_OP_NAMES


def is_top_level(ast_node, report=MAIN_REPORT) -> bool:
    """
    Determines if the `ast_node` is at the top-level of the program.
    Correctly handles expression statements (so a print call on its own will be
    considered a statement, even though its technically an expression).

    Args:
        ast_node (pedal.cait.cait_node.CaitNode): The CaitNode to check
        report (pedal.core.report.Report):

    Returns:
        bool: Whether the node is from the top level
    """
    ast = parse_program(report=report)
    for element in ast.body:
        if element.ast_name == "Expr":
            if element.value == ast_node:
                return True
        elif element == ast_node:
            return True
    return False


def find_operation(op_name, root=None, report=MAIN_REPORT):
    """
    Returns all the occurrences of the operator `op_name` in the source code.
    You can specify the operator as a string like `"+"` or `"<<"`.
    Supports all comparison, boolean, binary, and unary operators.
    """
    root = root or parse_program(report=report)
    found = []
    if op_name in COMPARE_OP_NAMES:
        compares = root.find_all("Compare")
        for compare in compares:
            for op in compare.ops:
                if op.ast_name == COMPARE_OP_NAMES[op_name]:
                    found.append(compare)
    elif op_name in BOOL_OP_NAMES:
        boolops = root.find_all("BoolOp")
        for boolop in boolops:
            if boolop.op_name == BOOL_OP_NAMES[op_name]:
                found.append(boolop)
    elif op_name in BIN_OP_NAMES:
        binops = root.find_all("BinOp")
        for binop in binops:
            if binop.op_name == BIN_OP_NAMES[op_name]:
                found.append(binop)
    elif op_name in UNARY_OP_NAMES:
        unaryops = root.find_all("UnaryOp")
        for unaryop in unaryops:
            if unaryop.op_name == UNARY_OP_NAMES[op_name]:
                found.append(unaryop)
    return found


def find_function_calls(name: str, root=None, report=MAIN_REPORT):
    """
    Returns a list of CaitNodes representing all of the function calls that
    were found. This includes both methods and regular functions.

    Args:
        name (str): The name of the function to search.
        root: A subtree of a parse tree to revert to.
        report (Report): The name of the Report to refer to.

    Returns:
        List[CaitNode]: Relevant call nodes.
    """
    root = root or parse_program(report=report)
    all_calls = root.find_all("Call")
    calls = []
    for a_call in all_calls:
        if a_call.func.ast_name == "Attribute":
            if a_call.func.attr == name:
                calls.append(a_call)
        elif a_call.func.ast_name == "Name":
            if a_call.func.id == name:
                calls.append(a_call)
    return calls


def function_is_called(name: str, report=MAIN_REPORT) -> int:
    """ DEPRECATED """
    return len(find_function_calls(name, report=report))


def find_prior_initializations(node, report=MAIN_REPORT):
    """
    DEPRECATED

    Given a Name node, returns a list of all the assignment
    statements that incorporate that Name node prior to that line. Returns
    None if no Name is given.

    """
    if node.ast_name != "Name":
        return None
    ast = parse_program(report=report)
    assignments = ast.find_all("Assign")
    cur_line_no = node.lineno
    all_assignments = []
    for assignment in assignments:
        if assignment.has(node):
            if assignment.lineno < cur_line_no:
                all_assignments.append(assignment)
    return all_assignments


def find_function_definition(name, root=None, report=MAIN_REPORT):
    """
    Finds the given function definition based on the given ``name``.

    Args:
        name (str): The name of the function.
        root: A subtree of a parse tree to revert to.
        report (Report): The name of the Report to refer to.

    Returns:
        :py:class:`pedal.cait.cait_node.CaitNode`: The first occurrence of a
            function with the given name.
    """
    root = root or parse_program(report=report)
    defs = root.find_all('FunctionDef')
    for a_def in defs:
        if a_def._name == name:
            return a_def
    return None



"""
Stores data class Identifier.
"""


class Identifier:
    """
    A representation of an Identifier, encapsulating its current level of
    existence, scope and State.

    Attributes:
        exists (bool): Whether or not the variable actually is defined anywhere.
                       It is possible that a variable was retrieved that does
                       not actually exist yet, which indicates it might need to
                       be created.
        in_scope (bool): Whether or not the variable exists in the current
                         scope. Used to detect the presence of certain kinds
                         of errors where the user is using a variable from
                         a different scope.
        scoped_name (str): The fully qualified name of the variable, including
                           its scope chain.
        state (State): The current state of the variable.
    """

    def __init__(self, exists: bool, in_scope: bool = False,
                 scoped_name: str = "UNKNOWN", state=""):
        self.exists = exists
        self.in_scope = in_scope
        self.scoped_name = scoped_name
        self.state = state



"""
A module that exposes a useful method (`timeout`) that can execute a
function asynchronously and terminiate if it exceeds a given `duration`.
"""

import sys
import time

try:
    import threading
except BaseException:
    threading = None
try:
    import ctypes
except BaseException:
    ctypes = None


class InterruptableThread(threading.Thread):
    """
    A thread that can be interrupted.
    """

    def __init__(self, func, args, kwargs):
        threading.Thread.__init__(self)
        self.func, self.args, self.kwargs = func, args, kwargs
        self.daemon = True
        self.result = None
        self.exc_info = (None, None, None)

    def run(self):
        """
        Begin thread execution, calling the `func` that was originally
        passed in.
        """
        try:
            self.result = self.func(*self.args, **self.kwargs)
        except Exception:
            self.exc_info = sys.exc_info()

    @staticmethod
    def _async_raise(thread_id, exception):
        '''
        Static method to raise an error asychronously using the ctypes module.
        '''
        # Cache the function for convenience
        RaiseAsyncException = ctypes.pythonapi.PyThreadState_SetAsyncExc

        states_modified = RaiseAsyncException(ctypes.c_long(thread_id),
                                              ctypes.py_object(exception))
        if states_modified == 0:
            raise ValueError("nonexistent thread id")
        elif states_modified > 1:
            RaiseAsyncException(thread_id, 0)
            raise SystemError("PyThreadState_SetAsyncExc failed")

    def raise_exception(self, exception):
        """
        Trigger a thread ending exception!
        """
        assert self.is_alive(), "thread must be started"
        for thread_id, thread in threading._active.items():
            if thread is self:
                InterruptableThread._async_raise(thread_id, exception)
                return

    def terminate(self):
        """

        """
        self.exc_info = sys.exc_info()
        self.raise_exception(SystemExit)


def timeout(duration, func, *args, **kwargs):
    """
    Executes a function and kills it (throwing an exception) if it runs for
    longer than the specified duration, in seconds.
    """

    # If libraries are not available, then we execute normally
    if None in (threading, ctypes):
        return func(*args, **kwargs)

    target_thread = InterruptableThread(func, args, kwargs)
    target_thread.start()
    target_thread.join(duration)

    if target_thread.is_alive():
        target_thread.terminate()
        timeout_exception = TimeoutError('Your code took too long to run '
                                         '(it was given {} seconds); '
                                         'maybe you have an infinite loop?'.format(duration))
        raise timeout_exception
    else:
        if target_thread.exc_info[0] is not None:
            ei = target_thread.exc_info
            # Python 2 had the three-argument raise statement; thanks to PEP
            # 3109 for showing how to convert that to valid Python 3 statements.
            e = ei[0](ei[1])
            e.__traceback__ = ei[2]
            e.exc_info = target_thread.exc_info
            raise e


# =========================================================================


class _TimeoutData:
    """
    Port of Craig Estep's AdaptiveTimeout JUnit rule from the VTCS student
    library.
    """

    # -------------------------------------------------------------
    def __init__(self, ceiling):
        self.ceiling = ceiling  # sec
        self.maximum = ceiling * 2  # sec
        self.minimum = 0.25  # sec
        self.threshold = 0.6
        self.rampup = 1.4
        self.rampdown = 0.5
        self.start = self.end = 0
        self.non_terminating_methods = 0

    # -------------------------------------------------------------

    def before_test(self):
        """
        Call this before a test case runs in order to reset the timer.
        """
        self.start = time.time()

    # -------------------------------------------------------------

    def after_test(self):
        """
        Call this after a test case runs. This will examine how long it took
        the test to execute, and if it required an amount of time greater than
        the current ceiling, it will adaptively adjust the allowed time for
        the next test.
        """
        self.end = time.time()
        diff = self.end - self.start

        if diff > self.ceiling:
            self.non_terminating_methods += 1

        if self.non_terminating_methods >= 2:
            if self.ceiling * self.rampdown < self.minimum:
                self.ceiling = self.minimum
            else:
                self.ceiling = (self.ceiling * self.rampdown)
        elif diff > self.ceiling * self.threshold:
            if self.ceiling * self.rampup > self.maximum:
                self.ceiling = self.maximum
            else:
                self.ceiling = (self.ceiling * self.rampup)



"""
Mock classes for open() and the file type.

The MIT License (MIT)

Copyright (c) 2015 Niv Ben-David

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import sys
from io import TextIOWrapper

from unittest.mock import Mock, NonCallableMock, DEFAULT

if sys.version_info < (3, 0):
    try:
        from cStringIO import StringIO, StringIO as BytesIO
    except ImportError:
        from StringIO import StringIO, StringIO as BytesIO
else:
    from io import StringIO, BytesIO


class FileLikeMock(NonCallableMock):
    """Acts like a file object returned from open()."""
    def __init__(self, name=None, read_data='', *args, **kws):
        kws.update({'spec': TextIOWrapper, })
        super(FileLikeMock, self).__init__(*args, **kws)
        self.mode = None
        self.__is_closed = False
        self.read_data = read_data
        self.close.side_effect = self._close
        self.__contents.seek(0)

        self.__enter__ = Mock(side_effect=self._enter)
        self.__exit__ = Mock(side_effect=self._exit)

        if name is not None:
            self.name = name

    @property
    def closed(self):
        # pylint: disable=missing-docstring
        return self.__is_closed

    @property
    def read_data(self):
        """Bypass read function to access the contents of the file.

        This property should be used for testing purposes.
        """
        return self.__contents.getvalue()

    @read_data.setter
    def read_data(self, contents):
        # pylint: disable=missing-docstring
        # pylint: disable=attribute-defined-outside-init
        if isinstance(contents, str):
            self.__contents = StringIO()
        else:
            self.__contents = BytesIO()

        # Constructing a cStrinIO object with the input string would result
        # in a read-only object, so we write the contents after construction.
        self.__contents.write(contents)

        # Set tell/read/write/etc side effects to access the new contents
        # object.
        self.tell._mock_wraps = self.__contents.tell
        self.seek._mock_wraps = self.__contents.seek
        self.read._mock_wraps = self.__contents.read
        self.readline._mock_wraps = self.__contents.readline
        self.readlines._mock_wraps = self.__contents.readlines
        self.write._mock_wraps = self.__contents.write
        self.writelines._mock_wraps = self.__contents.writelines

    def __iter__(self):
        return iter(self.__contents)

    def set_properties(self, path, mode):
        """Set file's properties (name and mode).

        This function is also in charge of swapping between textual and
        binary streams.
        """
        self.name = path
        self.mode = mode

        if 'b' in self.mode:
            if not isinstance(self.read_data, bytes):
                self.read_data = bytes(self.read_data, encoding='utf8')
        else:
            if not isinstance(self.read_data, str):
                self.read_data = str(self.read_data, encoding='utf8')

    def reset_mock(self, visited=None):
        """Reset the default tell/read/write/etc side effects."""
        # In some versions of the mock library, `reset_mock` takes an argument
        # and in some it doesn't. We try to handle all situations.
        if visited is not None:
            super(FileLikeMock, self).reset_mock(visited)
        else:
            super(FileLikeMock, self).reset_mock()

        # Reset contents and tell/read/write/close side effects.
        self.read_data = ''
        self.close.side_effect = self._close

    def _enter(self):
        """Reset the position in buffer whenever entering context."""
        self.__contents.seek(0)

        return self

    def _exit(self, exception_type, exception, traceback):
        """Close file when exiting context."""
        # pylint: disable=unused-argument
        self.close()

    def _close(self):
        """Mark file as closed (used for side_effect)."""
        self.__is_closed = True
        return DEFAULT


class MockOpen(Mock):
    """A mock for the open() builtin function."""
    def __init__(self, read_data='', *args, **kws):
        kws.update({'spec': open, 'name': open.__name__, })
        super(MockOpen, self).__init__(*args, **kws)
        self.__files = {}
        self.__read_data = read_data

    def __call__(self, path, mode='r', *args, **kws):
        original_side_effect = self._mock_side_effect

        if path in self.__files:
            self._mock_return_value = self.__files[path]
            self._mock_side_effect = self._mock_return_value.side_effect

        try:
            child = super(MockOpen, self).__call__(path, mode, *args, **kws)
        finally:
            # Reset the side effect after each call so that the next call to
            # open() won't cause the same side_effect.
            self._mock_side_effect = original_side_effect

        # Consecutive calls to open() set `return_value` to the last file mock
        # created. If the paths differ (and child isn't a newly-created mock,
        # evident by its name attribute being unset) we create a new file mock
        # instead of returning to previous one.
        # noinspection PyTypeHints
        if not isinstance(child.name, Mock) and path != child.name:
            child = self._get_child_mock(_new_name='()', name=path)
            self.__files[path] = child

        child.set_properties(path, mode)

        if path not in self.__files:
            self.__files[path] = child

        self._mock_return_value = child
        return child

    def __getitem__(self, path):
        return self.__files.setdefault(path, self._get_child_mock(name=path))

    def __setitem__(self, path, value):
        value.__enter__ = lambda self: self
        value.__exit__ = lambda self, *args: None
        self.__files[path] = value

    def reset_mock(self, visited=None):
        # See comment in `FileLikeMock.reset_mock`.
        if visited is not None:
            super(MockOpen, self).reset_mock(visited)
        else:
            super(MockOpen, self).reset_mock()

        self.__files = {}
        self.__read_data = ''

    def _get_child_mock(self, **kws):
        """Create a new FileLikeMock instance.

        The new mock will inherit the parent's side_effect and read_data
        attributes.
        """
        kws.update({
            '_new_parent': self,
            'side_effect': self._mock_side_effect,
            'read_data': self.__read_data,
        })
        return FileLikeMock(**kws)



"""
Collection of feedback functions related to static checks of student code.
"""
from pedal.assertions.functions import *
from pedal.cait.cait_api import parse_program
from pedal.assertions.feedbacks import AssertionFeedback
from pedal.cait.find_node import find_operation, find_function_calls, find_function_definition
from pedal.core.feedback import CompositeFeedbackFunction
from pedal.core.location import Location
from pedal.core.report import MAIN_REPORT
from pedal.core.commands import compliment as core_compliment, give_partial
from pedal.types.normalize import normalize_type
from pedal.types.operations import are_types_equal
from pedal.utilities.ast_tools import AST_NODE_NAMES


class EnsureAssertionFeedback(AssertionFeedback):
    """ Abstract base class for assertions preventing things. """
    def __init__(self, name, at_least=1, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        fields = {'name': name, 'at_least': at_least, 'capacity': '',
                  'root': root, 'name_message': report.format.name(name)}
        super().__init__(fields=fields, **kwargs)

    def _check_usage(self, field_name, uses):
        at_least = self.fields['at_least']
        self.fields[field_name] = use_count = len(uses)
        if at_least > use_count:
            if at_least == 1:
                return True
            else:
                self.fields['capacity'] = (f" at least {at_least} times, but"
                                           f" you used it {use_count} times")
                return True
        return False


class PreventAssertionFeedback(AssertionFeedback):
    """ Abstract base class for assertions preventing things. """
    def __init__(self, name, at_most=0, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        fields = {'name': name, 'at_most': at_most, 'capacity': '',
                  'root': root, 'name_message': report.format.name(name)}
        super().__init__(fields=fields, **kwargs)

    def _check_usage(self, field_name, uses):
        at_most = self.fields['at_most']
        self.fields[field_name] = use_count = len(uses)
        if use_count and at_most < use_count:
            if at_most == 0:
                return True
            else:
                self.fields['capacity'] = (f" more than {at_most} times, but"
                                           f" you used it {use_count} times")
                return True
        return False


class prevent_function_call(PreventAssertionFeedback):
    """
    Determine if a function was ever called by statically checking
    the students' code.

    # TODO: also check the TIFA records to see if the function was read

    Args:
        name (str): The name of the function (e.g., ``'sum'``)
        at_most (int): The maximum number of times you are allowed to call this
            function. Defaults to ``0``.
    """

    title = "May Not Use Function"
    message_template = ("You used the function {name_message} on line "
                        "{location.line}. You may not use that function"
                        "{capacity}.")

    def condition(self):
        """ Use find_matches to check number of occurrences. """
        name = self.fields['name']
        calls = find_function_calls(name, root=self.fields['root'])
        if calls:
            self.update_location(Location.from_ast(calls[-1]))
        return self._check_usage('call_count', calls)


class ensure_function_call(EnsureAssertionFeedback):
    """
    Determine that a function was definitely called by statically checking
    the students' code.

    # TODO: Allow parameter to force this to be a method xor/or function call
    # TODO: Allow set to be passed in for OR, or list for AND, for multiples

    # TODO: also check the TIFA records to see if the function was read

    Args:
        name (str): The name of the function (e.g., ``'sum'``)
        at_most (int): The maximum number of times you are allowed to call this
            function. Defaults to ``0``.
    """

    title = "Must Use Function"
    message_template = "You must use the function {name_message}{capacity}."

    def condition(self):
        """ Use find_matches to check number of occurrences. """
        name = self.fields['name']
        calls = find_function_calls(name, root=self.fields['root'])
        return self._check_usage('call_count', calls)


class prevent_operation(PreventAssertionFeedback):
    """
    Determines if the given operator `op_name` is not used anywhere, returning
    the node of it if it is. Otherwise, returns `None`. You can specify the
    operator as a string like `"+"` or `"<<"`. Supports all comparison,
    boolean, binary, and unary operators.

    Args:
        name (str): The name of the function (e.g., ``'sum'``)
        at_most (int): The maximum number of times you are allowed to call this
            function. Defaults to ``0``.
    """
    title = "May Not Use Operator"
    message_template = ("You used the operator {name_message} on line "
                        "{location.line}. You may not use that operator"
                        "{capacity}.")

    def condition(self):
        """ Use find_matches to check number of occurrences. """
        name = self.fields['name']
        self.fields['name_message'] = self.report.format.python_expression(name)
        uses = list(find_operation(name, root=self.fields['root']))
        if uses:
            self.update_location(Location.from_ast(uses[-1]))
        return self._check_usage('use_count', uses)


class ensure_operation(EnsureAssertionFeedback):
    """
    Determines if the given operator `op_name` is used somewhere. You can
    specify the operator as a string like `"+"` or `"<<"`. Supports all
    comparison, boolean, binary, and unary operators.

    Args:
        name (str): The name of the function (e.g., ``'sum'``)
        at_least (int): The minimum number of times you must call this
            function. Defaults to ``1``.
    """
    title = "Must Use Operator"
    message_template = "You must use the operator {name_message}{capacity}."

    def condition(self):
        """ Use find_matches to check number of occurrences. """
        name = self.fields['name']
        self.fields['name_message'] = self.report.format.python_expression(name)
        uses = list(find_operation(name, root=self.fields['root']))
        if uses:
            self.update_location(Location.from_ast(uses[-1]))
        return self._check_usage('use_count', uses)


class prevent_literal(PreventAssertionFeedback):
    """ Make sure that the given literal value does not appear in the student's
    code. """
    title = "May Not Use Literal Value"
    message_template = ("You used the literal value {literal_message:python_expression} on line "
                        "{location.line}. You may not use that value"
                        "{capacity}.")

    def __init__(self, literal, at_most=0, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        fields = {'literal': literal, 'at_most': at_most, 'capacity': '',
                  'root': root, 'literal_message': repr(literal)}
        super(AssertionFeedback, self).__init__(fields=fields, **kwargs)

    def condition(self):
        literal = self.fields['literal']
        uses = self.fields['root'].find_matches(repr(literal))
        if uses:
            self.update_location(uses[-1].match_location(False))
        return self._check_usage('use_count', uses)


class ensure_literal(EnsureAssertionFeedback):
    """ Make sure that the given literal value does appear in the student's
        code. """
    title = "Must Use Literal Value"
    message_template = "You must use the literal value {literal_message}{capacity}."

    def __init__(self, literal, at_least=1, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        fields = {'literal': literal, 'at_least': at_least, 'capacity': '',
                  'root': root, 'literal_message': report.format.python_expression(repr(literal))}
        super(AssertionFeedback, self).__init__(fields=fields, **kwargs)

    def condition(self):
        literal = self.fields['literal']
        uses = self.fields['root'].find_matches(repr(literal))
        if uses:
            self.update_location(uses[-1].match_location(False))
        return self._check_usage('use_count', uses)


class prevent_literal_type(PreventAssertionFeedback):
    """ Make sure that the given literal value does not appear in the student's
    code. """
    title = "May Not Use Type of Literal Value"
    message_template = ("You used the literal value type "
                        "{literal_type_message:python_expression} on line {location.line}."
                        " You may not use that type of value{capacity}.")

    def __init__(self, literal_type, at_most=0, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        fields = {'literal_type': literal_type, 'at_most': at_most,
                  'capacity': '', 'root': root,
                  'literal_type_message': literal_type.__name__}
        super(AssertionFeedback, self).__init__(fields=fields, **kwargs)

    def condition(self):
        literal_type = self.fields['literal_type']
        if literal_type == bool:
            uses = (self.fields['root'].find_matches("False") +
                    self.fields['root'].find_matches("True"))
            uses = [match.match_root for match in uses]
        elif literal_type == str:
            uses = self.fields['root'].find_all("Str")
        elif literal_type in (int, float):
            uses = [node for node in self.fields['root'].find_all("Num")
                    if isinstance(node.value, literal_type)]
        elif literal_type == list:
            uses = self.fields['root'].find_all("List")
        elif literal_type == dict:
            uses = self.fields['root'].find_all("Dict")
        else:
            raise ValueError(f"Unknown literal type: {literal_type}")
        if uses:
            self.update_location(uses[-1].lineno)
        return self._check_usage('use_count', uses)


class ensure_literal_type(EnsureAssertionFeedback):
    """ Make sure that the given type of literal value does appear in the
    student's code. """
    title = "Must Use Type of Literal Value"
    message_template = "You must use a literal value of type {literal_type_message:python_expression}{capacity}."

    def __init__(self, literal_type, at_least=1, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        fields = {'literal_type': literal_type, 'at_least': at_least,
                  'capacity': '', 'root': root,
                  'literal_type_message': literal_type.__name__}
        super(AssertionFeedback, self).__init__(fields=fields, **kwargs)

    def condition(self):
        literal_type = self.fields['literal_type']
        if literal_type == bool:
            uses = (self.fields['root'].find_matches("False")+
                    self.fields['root'].find_matches("True"))
            uses = [match.match_root for match in uses]
        elif literal_type == str:
            uses = self.fields['root'].find_all("Str")
        elif literal_type in (int, float):
            uses = [node for node in self.fields['root'].find_all("Num")
                    if isinstance(node.value, literal_type)]
        elif literal_type == list:
            uses = self.fields['root'].find_all("List")
        elif literal_type == dict:
            uses = self.fields['root'].find_all("Dict")
        else:
            raise ValueError(f"Unknown literal type: {literal_type}")
        if uses:
            self.update_location(uses[-1].lineno)
        return self._check_usage('use_count', uses)

class prevent_ast(PreventAssertionFeedback):
    """
    Determines if the given ast `name` is not used anywhere, returning
    the node of it if it is. You can refer to Green Tree Snakes documentation
    for more information about AST Names:

        https://greentreesnakes.readthedocs.io/en/latest/nodes.html

    Args:
        name (str): The name of the function (e.g., ``'sum'``)
        at_most (int): The maximum number of times you are allowed to use this
            AST. Defaults to ``0``.
    """
    title = "May Not Use Code"
    message_template = ("You used {name_message} on line "
                        "{location.line}. You may not use that"
                        "{capacity}.")

    def condition(self):
        """ Use find_all to check number of occurrences. """
        name = self.fields['name']
        self.fields['name_message'] = AST_NODE_NAMES.get(name, name)
        uses = list(self.fields['root'].find_all(name))
        if uses:
            self.update_location(Location.from_ast(uses[-1]))
        return self._check_usage('use_count', uses)


class ensure_ast(EnsureAssertionFeedback):
    """
    Determines if the given ast `name` is used anywhere, returning
    the node of it if it is. You can refer to Green Tree Snakes documentation
    for more information about AST Names:

        https://greentreesnakes.readthedocs.io/en/latest/nodes.html

    Args:
        name (str): The name of the node.
        at_least (int): The minimum number of times you must use this
            node. Defaults to ``1``.
    """
    title = "Must Use Code"
    message_template = "You must use {name_message}{capacity}."

    def condition(self):
        """ Use find_all to check number of occurrences. """
        name = self.fields['name']
        self.fields['name_message'] = AST_NODE_NAMES.get(name, name)
        uses = list(self.fields['root'].find_all(name))
        if uses:
            self.update_location(Location.from_ast(uses[-1]))
        return self._check_usage('use_count', uses)


def function_prints(function_name):
    """
    Determine that the print function is called within this function, by
    tracing its code statically. Does not actually verify that the function
    was ever called in practice.

    Args:
        function_name (str): The name of the function to search.

    Returns:

    """
    return ensure_function_call('print',
                                root=find_function_definition(function_name))


def has_import(ast, name):
    """
    Determine if the given module name is in the code.

    Args:
        ast (:py:class:`pedal.cait.cait_node.CaitNode`): The starting point
            to search.
        name (str): The name of the module to match against.

    Returns:
        bool: Whether or not the module name was matched.
    """
    imports = ast.find_all("Import")
    import_froms = ast.find_all("ImportFrom")
    # What about ``import <name>``
    if imports and any(alias._name == name for i in imports for alias in i.names):
        return True
    # What about ``from <name> import *``
    if import_froms and any(i.module == name for i in import_froms):
        return True
    return False


class ensure_import(EnsureAssertionFeedback):
    """
    Verify that the given ``module`` has been imported, by statically
    checking the import statements in the code. The ``at_least`` parameter
    is ignored.
    """
    title = "Must Import Module"
    message_template = "You must import the module {name_message}."

    def condition(self):
        """ Traverse the AST to find matches. """
        name = self.fields['name']
        ast = self.fields['root']
        return not has_import(ast, name)


class prevent_import(PreventAssertionFeedback):
    """
    Verify that the given ``module`` has not been imported, by statically
    checking the import statements in the code. The ``at_most`` parameter
    is ignored.
    """
    title = "May Not Import Module"
    message_template = "You may not import the module {name_message}."

    def condition(self):
        """ Traverse the AST to find matches. """
        name = self.fields['name']
        ast = self.fields['root']
        return has_import(ast, name)


class ensure_documented_functions(AssertionFeedback):
    """ Checks that all of the functions in the source code are documented. """
    title = "Must Document Functions"
    message_template = ("You must document the following function{plural}:"
                        " {names_message}.")

    def __init__(self, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        fields = {'root': root}
        super().__init__(fields=fields, **kwargs)

    def condition(self):
        """ Traverse the AST to find matches. """
        defs = self.fields['root'].find_all('FunctionDef')
        names = []
        for a_def in defs:
            # Don't have to document constructors (happens in Class Def)
            if a_def.name == "__init__":
                continue
            if (a_def.body and (a_def.body[0].ast_name != "Expr" or
                                a_def.body[0].value.ast_name not in ("Str", "Constant"))):
                names.append(a_def.name)
        self.fields['names'] = names
        self.fields['names_message'] = ", ".join(self.report.format.name(name)
                                                 for name in names)
        self.fields['plural'] = 's' if names else ''
        return bool(names)

"""
New assertions:

prevent_literal(literal_value: Any)
prevent_function_call(function_name: str) # Whether it was detected statically
prevent_operation(operator: str)
prevent_import(module_name: str)
prevent_ast(ast_name: str)

TODO: prevent_source_text(code: str, missing=False, minimum=2, maximum=3)
TODO: prevent_traced_call(function_name: str)# Whether it was detected dynamically
TODO: prevent_name(variable_name: str)
TODO: prevent_assignment(variable_name: str, type: str, value: Any)
TODO: ensure_function

ensure_prints(count)
    ensure_function_call('print', at_least=count)
    ensure_function_call('print', at_most=count)
function_prints => ensure_function_call('print', root=match_signature('name')

"""


# TODO: wait, is this a positive feedback group waiting to happen?

@CompositeFeedbackFunction(missing_function, duplicate_function_definition)
def ensure_function(name, arity=None, parameters=None,
                    returns=None, root=None, compliment=False, **kwargs):
    """ Checks that the function exists and has the right signature. """
    report = kwargs.get('report', MAIN_REPORT)
    root = root or parse_program(report=report)
    defs = root.find_all('FunctionDef')
    defs = [a_def for a_def in defs if a_def._name == name]
    if not defs:
        return missing_function(name, **kwargs)
    if len(defs) > 1:
        lines = [Location.from_ast(a_def) for a_def in defs]
        return duplicate_function_definition(name, lines, **kwargs)
    definition = defs[0]
    # Actual logic
    # 1.2.1 'arity' style - simply checks number of parameters
    if arity is not None or parameters is not None:
        expected_arity = arity if arity is not None else len(parameters)
        actual_arity = len(definition.args.args)
        if actual_arity < expected_arity:
            return too_few_parameters(name, actual_arity, expected_arity, **kwargs)
        elif actual_arity > expected_arity:
            return too_many_parameters(name, actual_arity, expected_arity, **kwargs)
    # 1.2.2 'parameters' style - checks each parameter's name and type
    if parameters is not None:
        actual_parameters = definition.args.args
        for expected_parameter, actual_parameter in zip(parameters, actual_parameters):
            expected_parameter_type = normalize_type(expected_parameter)
            actual_parameter_name = (actual_parameter.id if actual_parameter.id is not None
                                     else actual_parameter.arg)
            if actual_parameter.annotation is None:
                return missing_parameter_type(name, actual_parameter_name,
                                              expected_parameter_type,
                                              **kwargs)
            try:
                actual_parameter_type = normalize_type(actual_parameter.annotation.ast_node)
            except ValueError as e:
                return invalid_parameter_type(name, actual_parameter_name,
                                              actual_parameter.annotation,
                                              expected_parameter_type,
                                              **kwargs)
            if not are_types_equal(actual_parameter_type, expected_parameter_type):
                return wrong_parameter_type(name, actual_parameter_name,
                                            actual_parameter_type,
                                            expected_parameter_type, **kwargs)
    # 1.2.3. 'returns' style - checks the return type explicitly
    if returns is not None:
        expected_returns = normalize_type(returns)
        if definition.returns is None:
            return missing_return_type(name, expected_returns, **kwargs)
        try:
            actual_returns = normalize_type(definition.returns.ast_node)
        except ValueError as e:
            return invalid_return_type(name, definition.returns,
                                       expected_returns, **kwargs)
        if not are_types_equal(actual_returns, expected_returns):
            return wrong_return_type(name, actual_returns, expected_returns,
                                     **kwargs)
    # Alternatively, returns positive FF?
    if compliment:
        if isinstance(compliment, str):
            core_compliment(compliment, label="function_defined", **kwargs)
        elif compliment is True:
            core_compliment(f"Defined {name}", label="function_defined", **kwargs)
    elif kwargs.get("score"):
        give_partial(kwargs.pop("score"), label="function_defined", **kwargs)

    return None


@CompositeFeedbackFunction(prevent_function_call, ensure_function_call)
def ensure_prints_exactly(count, **kwargs):
    """
    DEPRECATED.

    Used to ensure that the student's code calls the print function a certain
    number of times.

    Args:
        count (int): The minimum and maximum number of times you can call print.

    Returns:
        bool: Whether or not both the maximum and minimum was met.
    """
    return (ensure_function_call('print', at_least=count, **kwargs) and
            prevent_function_call('print', at_most=count, **kwargs))


class ensure_starting_code(AssertionFeedback):
    """Detects if the given code is missing."""
    title = "Don't Change Starting Code"
    message_template = "You have changed or removed the starting code. "

    def __init__(self, code, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        fields = {'code': code, 'root': root}
        super().__init__(fields=fields, **kwargs)

    def condition(self):
        """ Traverse the AST to find matches. """
        code = self.fields['code']
        ast = self.fields['root']
        return not ast.find_match(code)


class prevent_embedded_answer(AssertionFeedback):
    """ Detects if the given code is present. """
    title = "Don't Write Answer Directly"
    message_template = ("You have embedded the answer directly in your code, "
                        "instead of writing code to compute the answer.")

    def __init__(self, code, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        fields = {'code': code, 'root': root}
        super().__init__(fields=fields, **kwargs)

    def condition(self):
        """ Traverse the AST to find matches. """
        code = self.fields['code']
        ast = self.fields['root']
        return ast.find_match(code)



"""
Utility function for taking various kinds of "type expressions" and turning
them into a canonical Pedal type.

* Pedal Type
* Python builtin type (e.g., `list`, `int`)
    get_pedal_type_from_builtin
* Python literal value (e.g., `5`, `"Hello World!"`)
    get_pedal_type_from_value
* AST Nodes (e.g., `ast.Num`)
    get_pedal_type_from_ast
* JSON-Encoded representation of a type (e.g., from CORGIS datasets)
    get_pedal_type_from_json
* String representation of parseable Python code
    get_pedal_type_from_str
* Python Typing module type (e.g., `List`, `Generic`)
    TODO: Implement get_pedal_type_from_typing
"""
import ast
from pedal.types.definitions import (Type, UnknownType, NumType, BoolType,
                                     TupleType, ListType, StrType,
                                     DictType, SetType, GeneratorType,
                                     ModuleType, NoneType, FrozenSetType,
                                     FunctionType, TYPE_STRINGS,
                                     LiteralStr, LiteralNum, LiteralBool, CustomType)
from pedal.utilities.system import IS_PYTHON_39


def normalize_type(type_expression, type_space=None):
    """
    Converts the given ``type_expression`` into a normalized Pedal Type.

    Args:
        type_expression (Any): Any kind of supported "Type Expression". This
            includes builtin Python types, string representations, AST nodes,
            JSON-Encoded representations.

    Returns:
        :py:class:`pedal.types.definitions.Type`: A normalized Pedal Type.
    """
    # Already a Pedal Type
    if isinstance(type_expression, Type):
        return type_expression
    # What if it's a builtin type function?
    if isinstance(type_expression, type):
        return normalize_type(type_expression.__name__, type_space=type_space)
    # Might be a string, can we evaluate?
    if isinstance(type_expression, str):
        result = get_pedal_type_from_str(type_expression, type_space)
        if not isinstance(result, (UnknownType, CustomType)):
            return result
        try:
            first_line = ast.parse(type_expression).body[0].value
        except SyntaxError as e:
            raise ValueError(f"Could not parse string {type_expression!r} into "
                             f"a Pedal type:\n {e}")
        except AttributeError as e:
            raise ValueError(f"Could not parse string {type_expression!r} into "
                             f"a Pedal type:\n This was not an expression.")
        try:
            return normalize_type(first_line, type_space=type_space)
        except:
            return result
    # Might be a JSON-Encoded dictionary
    if isinstance(type_expression, dict):
        try:
            return get_pedal_type_from_json(type_expression)
        except KeyError:
            return get_pedal_type_from_type_literal(type_expression, type_space)
    # Perhaps it is an AST node
    if isinstance(type_expression, ast.AST):
        return get_pedal_type_from_ast(type_expression, type_space=type_space)
    if isinstance(type_expression, list):
        return get_pedal_type_from_type_literal(type_expression, type_space)
    raise ValueError(f"Could not normalize {type_expression!r} into a Pedal type.")


def get_pedal_type_from_json(val):
    """

    Args:
        val:

    Returns:

    """
    if val['type'] == 'DictType':
        values = [get_pedal_type_from_json(v) for v in val['values']]
        empty = val.get('empty', None)
        if 'literals' in val:
            literals = [get_pedal_literal_from_json(l) for l in val['literals']]
            return DictType(empty, literals=literals, values=values)
        else:
            keys = [get_pedal_type_from_json(k) for k in val['keys']]
            return DictType(empty, keys=keys, values=values)
    elif val['type'] == 'ListType':
        return ListType(get_pedal_type_from_json(val.get('subtype', None)),
                        val.get('empty', None))
    elif val['type'] == 'StrType':
        return StrType(val.get('empty', None))
    elif val['type'] == 'BoolType':
        return BoolType()
    elif val['type'] == 'NoneType':
        return NoneType()
    elif val['type'] == 'NumType':
        return NumType()
    elif val['type'] == 'ModuleType':
        submodules = {name: get_pedal_type_from_json(m)
                      for name, m in val.get('submodules', {}).items()}
        fields = {name: get_pedal_type_from_json(m)
                  for name, m in val.get('fields', {}).items()}
        return ModuleType(name=val.get('name'), submodules=submodules,
                          fields=fields)
    elif val['type'] == 'FunctionType':
        returns = get_pedal_type_from_json(val.get('returns', {'type': 'NoneType'}))
        return FunctionType(name=val.get('name'), returns=returns)


def get_pedal_literal_from_json(val):
    """

    Args:
        val:

    Returns:

    """
    if val['type'] == 'LiteralStr':
        return LiteralStr(val['value'])
    elif val['type'] == 'LiteralNum':
        return LiteralNum(val['value'])
    elif val['type'] == 'LiteralBool':
        return LiteralBool(val['value'])


def get_pedal_literal_from_pedal_type(type):
    """

    Args:
        type:

    Returns:

    """
    if isinstance(type, NumType):
        return LiteralNum(0)
    elif isinstance(type, StrType):
        return LiteralStr("")
    else:
        # TODO: Finish the mapping
        return LiteralStr("")


def get_pedal_type_from_str(value, tifa_instance=None):
    """

    Args:
        value:
        tifa_instance:

    Returns:

    """
    # if value in custom_types:
    #    return custom_types[value]
    if value.lower() in TYPE_STRINGS:
        return TYPE_STRINGS[value.lower()]()
    elif tifa_instance:
        variable = tifa_instance.find_variable_scope(value)
        if variable.exists:
            state = tifa_instance.load_variable(value)
            return state.type
        # custom_types.add(value)
    return CustomType(value)
    # TODO: handle custom types


def get_pedal_type_from_annotation(v, tifa_instance=None):
    """

    Attempts to intelligently parse the type, which might be a variable or
    a string node.

    Args:
        v:
        tifa_instance:

    Returns:

    """
    if isinstance(v, ast.Str):
        return get_pedal_type_from_str(v.s, tifa_instance)
    elif isinstance(v, ast.Name):
        return get_pedal_type_from_str(v.id, tifa_instance)
    # Fall back to default AST behavior
    return get_pedal_type_from_ast(v, tifa_instance)


def get_pedal_type_from_value(value, type_space=None) -> Type:
    """ Converts the Python value to a Pedal Type """
    if isinstance(value, bool):
        return BoolType()
    if isinstance(value, (int, float, complex)):
        return NumType()
    if isinstance(value, str):
        return StrType()
    if isinstance(value, type(None)):
        return NoneType()
    if isinstance(value, tuple):
        return TupleType((get_pedal_type_from_value(t, type_space) for t in value))
    if isinstance(value, frozenset):
        return FrozenSetType((get_pedal_type_from_value(t, type_space) for t in value))
    if isinstance(value, list):
        if value:
            return ListType(empty=False, subtype=get_pedal_type_from_value(value[0]))
        else:
            return ListType(empty=True)
    if isinstance(value, dict):
        if not value:
            return DictType(empty=True)
        if all(isinstance(value, str) for k in value.keys()):
            return DictType(literals=[LiteralStr(s) for s in value.keys()],
                            values=[get_pedal_type_from_value(vv, type_space)
                                    for vv in value.values()])
        return DictType(keys=[get_pedal_type_from_ast(k, type_space)
                              for k in value.keys()],
                        values=[get_pedal_type_from_ast(vv, type_space)
                                for vv in value.values()])
    return UnknownType()


def get_pedal_type_from_ast(value: ast.AST, type_space=None) -> Type:
    """
    Determines the Pedal Type from this ast node.
    Args:
        value (ast.AST): An AST node.

    Returns:
        Type: A Pedal Type
    """
    try:
        if isinstance(value, ast.Constant):
            return get_pedal_type_from_value(value.value, type_space)
    except AttributeError as e:
        pass
    if isinstance(value, ast.Name):
        return get_pedal_type_from_str(value.id, type_space)
    elif isinstance(value, ast.Str):
        return StrType(bool(value.v))
    elif isinstance(value, ast.List):
        return ListType(subtype=(get_pedal_type_from_ast(value.elts[0], type_space)
                                 if value.elts else None),
                        empty=not bool(value.elts))
    elif isinstance(value, ast.Set):
        return SetType(subtype=(get_pedal_type_from_ast(value.elts[0], type_space)
                                if value.elts else None),
                       empty=not bool(value.elts))
    elif isinstance(value, ast.Tuple):
        return TupleType(subtypes=[get_pedal_type_from_ast(e, type_space)
                                   for e in value.elts])
    elif isinstance(value, ast.Dict):
        if not value.keys:
            return DictType(empty=True)
        if all(isinstance(k, ast.Str) for k in value.keys):
            return DictType(literals=[LiteralStr(s.s) for s in value.keys],
                            values=[get_pedal_type_from_ast(vv, type_space)
                                    for vv in value.values])
        return DictType(keys=[get_pedal_type_from_ast(k, type_space)
                              for k in value.keys],
                        values=[get_pedal_type_from_ast(vv, type_space)
                                for vv in value.values])
    # Support new style subscripts (e.g., ``list[int]``)
    elif ((IS_PYTHON_39 and isinstance(value, ast.Subscript)) or
          isinstance(value, ast.Subscript) and isinstance(value.slice, ast.Index)):
        if IS_PYTHON_39:
            slice = value.slice
        else:
            slice = value.slice.value
        if isinstance(value.value, ast.Name):
            if isinstance(slice, ast.Name):
                subtype = get_pedal_type_from_str(slice.id)
                if value.value.id == "list":
                    return ListType(subtype=subtype, empty=False)
                if value.value.id == "set":
                    return SetType(subtype=subtype, empty=False)
                if value.value.id == "tuple":
                    return TupleType(subtypes=(subtype,))
                if value.value.id == "frozenset":
                    return FrozenSetType(subtype=subtype, empty=False)
            elif isinstance(slice, ast.Tuple):
                subtypes = [get_pedal_type_from_ast(e, type_space)
                            for e in slice.elts]
                if value.value.id == "tuple":
                    return TupleType(subtypes=subtypes)
                elif value.value.id == "dict" and len(subtypes) == 2:
                    return DictType(keys=subtypes[0], values=subtypes[1])
    # Top-level Module, parse it and get it back
    if isinstance(value, ast.Module) and value.body:
        if isinstance(value.body[0], ast.Expr):
            return get_pedal_type_from_ast(value.body[0].value, type_space)
    return UnknownType()


def get_pedal_type_from_type_literal(value, type_space=None) -> Type:
    if value == list:
        return ListType(empty=False)
    elif value == dict:
        return DictType(empty=False)
    elif isinstance(value, list):
        if value:
            return ListType(empty=False, subtype=get_pedal_type_from_type_literal(value[0], type_space))
        else:
            return ListType(empty=True)
    elif isinstance(value, dict):
        if not value:
            return DictType(empty=True)
        if all(isinstance(value, str) for k in value.keys()):
            return DictType(literals=[LiteralStr(s) for s in value.keys()],
                            values=[get_pedal_type_from_type_literal(vv, type_space)
                                    for vv in value.values()])
        return DictType(keys=[get_pedal_type_from_type_literal(k, type_space)
                              for k in value.keys()],
                        values=[get_pedal_type_from_type_literal(vv, type_space)
                                for vv in value.values()])
    elif isinstance(value, type):
        return get_pedal_type_from_str(value.__name__)
    else:
        return get_pedal_type_from_str(str(value))



"""
A package for analyzing student code.
"""
import sys
import os

# Core Commands
from pedal.core.report import MAIN_REPORT
from pedal.core.submission import Submission
from pedal.core.commands import *
from pedal.source import *
from pedal.sandbox.commands import *
from pedal.cait import *
from pedal.assertions.commands import *

student: Sandbox



"""
Utilities for working with Abstract Syntax Trees.
"""
import ast

__all__ = ['is_literal', 'FindExecutableLines']

AST_NODE_NAMES = {
    "Add": "an addition operator",
    "And": "a boolean AND operator",
    "AnnAssign": "an annotated assignment",
    "Assert": "an assert statement",
    "Assign": "an assignment statement",
    "AsyncFor": "an asychronous for loop",
    "AsyncFunctionDef": "an asychronous function definition",
    "AsyncWith": "an asychronous with statement",
    "Attribute": "an attribute lookup",
    "AugAssign": "an augmented assignment",
    "AugLoad": "an augmented load",
    "AugStore": "an augmented store",
    "Await": "an await statement",
    "BinOp": "a binary operator",
    "BitAnd": "a bitwise AND operator",
    "BitOr": "a bitwise OR operator",
    "BitXor": "a bitwise XOR operator",
    "BoolOp": "a boolean operator",
    "Break": "a break statement",
    "Bytes": "a literal bytes string",
    "Call": "a function call",
    "ClassDef": "a class definition",
    "Compare": "a boolean comparison",
    "Constant": "a literal value",
    "Continue": "a continue statement",
    "Del": "a delete statement",
    "Delete": "a deletion",
    "Dict": "a dictionary literal",
    "DictComp": "a dictionary comprehension",
    "Div": "a division operator",
    "Ellipsis": "an ellipsis",
    "Eq": "an equality comparison operator",
    "ExceptHandler": "an except handler",
    "Expr": "an expression used as a statement",
    "Expression": "an evaluated expression",
    "ExtSlice": "a multi-dimensional slice",
    "FloorDiv": "an integer division operator",
    "For": "a FOR loop",
    "FormattedValue": "a formatted value in an f-string",
    "FunctionDef": "a function definition",
    "GeneratorExp": "a generator expression",
    "Global": "a global statement",
    "Gt": "a greater than comparison operator",
    "GtE": "a greater than or equal to comparison operator",
    "If": "an IF statement",
    "IfExp": "an IF expression",
    "Import": "an import statement",
    "ImportFrom": "an import/from statement",
    "In": "an IN operator",
    "Index": "an index",
    "Interactive": "an interactive expression",
    "Invert": "an invert operator",
    "Is": "an IS operator",
    "IsNot": "an IS NOT operator",
    "JoinedStr": "an f-string",
    "LShift": "a left shift operator",
    "Lambda": "a lambda expression",
    "List": "a list literal",
    "ListComp": "a list comprehension",
    "Load": "a load",
    "Lt": "a less than comparison operator",
    "LtE": "a less than or equal to comparison operator",
    "MatMult": "a matrix multiplication operator",
    "Mod": "a modulo operator",
    "Module": "a module",
    "Mult": "a multiplication operator",
    "Name": "a name",
    "NameConstant": "a name constant",
    "Nonlocal": "a nonlocal statement",
    "Not": "a not operator",
    "NotEq": "a not equal to comparison operator",
    "NotIn": "a NOT IN operator",
    "Num": "a numeric literal",
    "Or": "a boolean OR operator",
    "Param": "a parameter",
    "Pass": "a pass statement",
    "Pow": "a power operator",
    "RShift": "a right shift operator",
    "Raise": "a raise statement",
    "Return": "a return statement",
    "Set": "a set literal",
    "SetComp": "a set comprehension",
    "Slice": "a slice",
    "Starred": "a starred argument",
    "Store": "a store",
    "Str": "a string literal",
    "Sub": "a subtraction operator",
    "Subscript": "a subscript",
    "Suite": "a suite",
    "Try": "a try statement",
    "Tuple": "a tuple literal",
    "TypeIgnore": " a type ignore",
    "UAdd": "a unary addition operator",
    "USub": "a unary subtraction operator",
    "UnaryOp": "a unary operator",
    "While": "a while loop",
    "With": "a with statement",
    "Yield": "a yield statement",
    "YieldFrom": "a yield/from statement"
}


def is_literal(node: ast.AST) -> bool:
    """
    Consumes an AST Node and produces whether or not it is a literal value.
    Args:
        node (AST): The AST node to check

    Returns:
        bool: Whether or not this is an AST Node
    """
    # Simple primitive constant?
    if isinstance(node, (ast.Num, ast.Str, ast.Bytes, ast.Ellipsis,
                         ast.NameConstant)):
        return True
    # V3.8 Constant?
    try:
        if isinstance(node, ast.Constant):
            return True
    except AttributeError:
        pass
    # V3.6 FormattedValue?
    try:
        if isinstance(node, ast.FormattedValue):
            return True
    except AttributeError:
        pass
    # Iterable elts?
    if isinstance(node, (ast.List, ast.Tuple, ast.Set)):
        return all(is_literal(elt) for elt in node.elts)
    # Dictionary?
    if isinstance(node, ast.Dict):
        return all(is_literal(key) and is_literal(value)
                   for key, value in zip(node.keys, node.values))
    # Nope, not a literal
    return False


class FindExecutableLines(ast.NodeVisitor):
    """
    NodeVisitor subclass that visits every statement of a program and tracks
    their line numbers in a list.

    Attributes:
        lines (list[int]): The list of lines that were visited.
    """

    def __init__(self):
        self.lines = []

    def _track_lines(self, node):
        self.lines.append(node.lineno)
        self.generic_visit(node)

    def _track_expr_line(self, node):
        """ Skips over string Expr nodes (which are docstrings) """
        try:
            node.value.s
        except AttributeError:
            self._track_lines(node)

    visit_FunctionDef = _track_lines
    visit_AsyncFunctionDef = _track_lines
    visit_ClassDef = _track_lines
    visit_Return = _track_lines
    visit_Delete = _track_lines
    visit_Assign = _track_lines
    visit_AugAssign = _track_lines
    visit_AnnAssign = _track_lines
    visit_For = _track_lines
    visit_AsyncFor = _track_lines
    visit_While = _track_lines
    visit_If = _track_lines
    visit_With = _track_lines
    visit_AsyncWith = _track_lines
    visit_Raise = _track_lines
    visit_Try = _track_lines
    visit_Assert = _track_lines
    visit_Import = _track_lines
    visit_ImportFrom = _track_lines
    visit_Global = _track_lines
    visit_Nonlocal = _track_lines
    visit_Expr = _track_expr_line
    visit_Pass = _track_lines
    visit_Continue = _track_lines
    visit_Break = _track_lines



"""
The Tag class.
"""


class Tag:
    """
    Tags are used to store additional metadata for a piece of feedback
    in a semi-structured way.

    TODO: These currently are just placeholders until we make up a design!

    """
    def __init__(self, kind, name, explanation, weight):
        self.kind = kind
        self.name = name
        self.explanation = explanation
        self.weight = weight



"""

Resolver Types

Does there need to be some kind of hook for Tools to wrap up their business?

Simple
    Find the highest priority feedback and show that, along with any positive feedback.
    Break ties by showing the first one that was triggered.

Sectional
    Find the highest priority feedback for each section, and show that along with any positive feedback.

Full
    Report all feedback, grouped by tool/category/priority/time.

Full Summary
    Report all feedback but divided into frequencies of labels grouped by tool/category/priority/time.

"""

from pedal.resolvers import simple


def print_resolve(*args, **kwargs):
    """
    Trivial formatter for resolver, just dumps the
    Title/Label/Score/Message. Any arguments are forwarded to
    :py:func:`pedal.resolvers.simple.resolve`
    """
    result = simple.resolve(*args, **kwargs)
    print("Title:", result.title)
    print("Label:", result.label)
    print("Score:", result.score)
    print("Message:", result.message)
    return result



from pedal.core.commands import gently
from pedal.core.report import MAIN_REPORT
from pedal.questions import _name_hash, SETTING_SHOW_CASE_DETAILS, load_question
from pedal.questions.questions import show_question


class Pool:
    """

    """
    _POOL_TRACKER = 0

    def __init__(self, name, choices, seed=None, report=MAIN_REPORT, position=None):
        self.name = name
        self.choices = choices
        self.seed = seed
        self.report = report
        if position is None:
            position = Pool._POOL_TRACKER
            Pool._POOL_TRACKER += 1
        self.position = position

    def choose(self, force=None):
        """

        Args:
            force:

        Returns:

        """
        if force is None:
            if self.seed is None:
                force = self.report['questions']['seed']
                if isinstance(force, str):
                    force = _name_hash(force + self.name)
                # Assume iterable; could be check that throws better error
                if not isinstance(force, int):
                    force = force[self.position]
            else:
                force = self.seed
        return self.choices[force % len(self.choices)]

    @property
    def answered(self):
        """

        Returns:

        """
        for choice in self.choices:
            if choice.answered:
                return True
        return False


def check_pool_exam(name, questions, force=None, seed=None, report=MAIN_REPORT):
    """

    Args:
        name:
        questions:
        force:
        seed:
        report:
    """
    # Choose a question
    if force is None:
        if seed is None:
            force = MAIN_REPORT['questions']['seed']
            if isinstance(force, str):
                force = _name_hash(force + name)
        else:
            force = seed
    elif isinstance(force, str):
        force = _name_hash(force + name)
    question = questions[force % len(questions)]
    # Ask it
    show_question(question['instructions'])
    # Check if they're done
    if 'settings' not in question:
        question['settings'] = {}
    question['settings'][SETTING_SHOW_CASE_DETAILS] = False
    results = list(load_question(question))
    if results:
        message, label = results[0]
        gently(message, label=label)



"""
Extensions related to built-in Turtle library.

Mock Module for Sandbox.
"""

from pedal.types.builtin import BUILTIN_MODULES
from pedal.types.definitions import ModuleType, FunctionType, NoneType

_TURTLE_FIELDS = {
   'forward': FunctionType(name='forward', returns=NoneType()),
   'backward': FunctionType(name='backward', returns=NoneType()),
   'color': FunctionType(name='color', returns=NoneType()),
   'right': FunctionType(name='right', returns=NoneType()),
   'left': FunctionType(name='left', returns=NoneType()),
}
BUILTIN_MODULES['turtle'] = ModuleType('turtle',
                                       fields=_TURTLE_FIELDS),



"""
Extensions related to MatPlotLib.

Mock module for Sandbox.

Assertions.
"""
from pedal import Feedback, CompositeFeedbackFunction
from pedal.assertions import ensure_function_call, prevent_function_call
from pedal.core.feedback import FeedbackResponse
from pedal.core.report import MAIN_REPORT
from pedal.types.builtin import BUILTIN_MODULES
from pedal.types.definitions import ModuleType, FunctionType, NoneType
from pedal.cait.find_node import function_is_called
from pedal.cait.cait_api import parse_program, def_use_error
from pedal.sandbox.commands import get_sandbox

PLOT_LABEL = {'plot': 'line plot',
              'hist': 'histogram',
              'scatter': 'scatter plot'}


class plt_rename_err(Feedback):
    title = "Wrong MatPlotLib Import"
    priority = Feedback.CATEGORIES.SYNTAX
    category = Feedback.CATEGORIES.INSTRUCTOR
    justification = "The name 'plt' appeared in the code with a def-use error."
    constant_fields = {'suggestion': 'import matplotlib.pyplot as plt',
                       'actual': 'matplotlib.pyplot',
                       'expected': 'plt'}
    message_template = ("You have imported the "
                        "{expected:name} module, "
                        "but you did not rename it to {actual:name} "
                        "using {suggestion:python_expression}.")

    def condition(self):
        ast = parse_program(report=self.report)
        plts = [n for n in ast.find_all("Name") if n.id == 'plt']
        if plts and any(def_use_error(plt) for plt in plts):
            return True
        return False


class plt_wrong_import(Feedback):
    title = "Missing MatPlotLib Import"
    priority = Feedback.CATEGORIES.SYNTAX
    category = Feedback.CATEGORIES.INSTRUCTOR
    justification = ("A matplotlib name (e.g., 'plot' or 'hist') was used with"
                     " a def-use error.")
    constant_fields = {'suggestion': 'import matplotlib.pyplot as plt'}
    message_template = ("You have attempted to use the MatPlotLib "
                        "function named {expected:name}. However, you "
                        "imported MatPlotLib in a way that does not "
                        "allow you to use the function directly. I "
                        "recommend you use {actual:python_expression} instead, "
                        "after you use {suggestion:python_expression}.")

    def condition(self):
        ast = parse_program(report=self.report)
        matplotlib_names = ['plot', 'hist', 'scatter',
                            'title', 'xlabel', 'ylabel', 'show']
        for name in matplotlib_names:
            for n in ast.find_all("Name"):
                if n.id == name:
                    if def_use_error(n):
                        self.fields['actual'] = name
                        self.fields['expected'] = "plt." + name
                        return True
        return False


@CompositeFeedbackFunction(plt_rename_err, plt_wrong_import)
def prevent_incorrect_plt(**kwargs):
    """ Confirms that matplotlib.pyplot is being imported correctly. """
    return plt_rename_err(**kwargs) or plt_wrong_import(**kwargs)


@CompositeFeedbackFunction(prevent_function_call, ensure_function_call)
def ensure_correct_plot(function_name, report=MAIN_REPORT, **kwargs):
    """ Checks that the given plot type was correctly called. """
    ensure_function_call(function_name, **kwargs)
    as_code = report.format.python_expression
    for name, description in PLOT_LABEL.items():
        if name == function_name:
            continue
        prevent_function_call(name, **kwargs,
                              message=(f"You have called the {as_code(name)} "
                                       f"function, which makes a {description}."))


class ensure_show(Feedback):
    """ Verifies that the `plt.show` function was called. """
    title = "Missing Show Function"
    category = Feedback.CATEGORIES.INSTRUCTOR
    constant_fields = {'missing': 'plt.show'}
    message_template = ("You have not called the {missing:python_expression} "
                        "function, which actually creates the graph.")
    justification = "The show function was not found as a function call."

    def condition(self):
        return not function_is_called("show")


def compare_data(plt_type, correct, given, special_comparion=None):
    """
    Determines whether the given data matches any of the data found in the
    correct data. This handles plots of different types: if a histogram
    was plotted with the expected data for a line plot, it will return True.

    Args:
        plt_type (str): The expected type of this plot
        correct (List of Int or List of List of Int): The expected data.
        given (Dict): The actual plotted data and information
        special_comparion (callable): A special comparison function to use
            between the data. If None, then will use the ``==`` operator.
    Returns:
        bool: Whether the correct data was found in the given plot.
    """
    if special_comparion is None:
        def special_comparion(left, right):
            return left == right

    # Infer arguments
    if plt_type == 'hist':
        correct_xs = None
        correct_ys = correct
    elif not correct:
        correct_xs = []
        correct_ys = []
    elif isinstance(correct[0], (tuple, list)):
        # We were given a list of lists of ints
        correct_xs, correct_ys = correct
    else:
        # Assume it is a singular list
        correct_xs = list(range(len(correct)))
        correct_ys = correct

    if given['type'] == 'hist':
        return special_comparion(correct_ys, given['values'])
    elif plt_type == 'hist':
        return special_comparion(correct_ys, given['y'])
    else:
        return (special_comparion(correct_xs, given['x']) and
                special_comparion(correct_ys, given['y']))


GRAPH_TYPES = {'line': 'line plot',
               'hist': 'histogram',
               'scatter': 'scatter plot'}


class BadGraphFeedback(FeedbackResponse):
    category = Feedback.CATEGORIES.INSTRUCTOR
    def __init__(self, plt_type, expected, actual, **kwargs):
        fields = kwargs.setdefault('fields', {})
        fields['plt_type'] = plt_type
        fields['expected'] = expected
        fields['actual'] = actual
        super().__init__(plt_type, expected, actual, **kwargs)


class other_plt(BadGraphFeedback):
    title = "Plotting Another Graph"
    message_template = ("You have created a {plt_type}, but it does not "
                        "have the right data. That data appears to have been "
                        "plotted in another graph.")


class wrong_plt_data(BadGraphFeedback):
    title = "Plot Data Incorrect"
    message_template = ("You have created a {plt_type}, but it does not have "
                        "the right data.")


class wrong_plt_type(BadGraphFeedback):
    title = "Wrong Plot Type"
    message_template = ("You have plotted the right data, but you appear to "
                        "have not plotted it as a {plt_type}.")


class no_plt(BadGraphFeedback):
    title = "Missing Plot"
    message_template = "You have not created a {plt_type} with the proper data."


@CompositeFeedbackFunction(other_plt, wrong_plt_data, wrong_plt_type, no_plt)
def assert_plot(plt_type, data, **kwargs):
    """
    Check whether a plot with the given ``plt_type`` and ``data`` exists.
    If the plot was found successfully, returns False.
    Otherwise, returns the feedback that was detected.

    Args:
        plt_type (str): Either 'line', 'hist', or 'scatter'
        data (list): The expected data to check in the plots. Could be a single
            list of numbers, or a pair of two lists (for scatter/line plots).
    """
    report = kwargs.get("report", MAIN_REPORT)
    # Allow instructor to use "plot" instead of "line" as type
    if plt_type == 'plot':
        plt_type = 'line'
    # Check the plots to see if there is a plot with the data
    type_found = False
    data_found = False
    plots = get_sandbox(report=report).modules.plotting.plots
    for graph in plots:
        for a_plot in graph['data']:
            data_found_here = compare_data(plt_type, data, a_plot)
            if a_plot['type'] == plt_type and data_found_here:
                return False
            if a_plot['type'] == plt_type:
                type_found = True
            if data_found_here:
                data_found = data_found_here
    # Figure out what kind of mistake was made.
    plt_type = GRAPH_TYPES.get(plt_type, plt_type)
    if type_found and data_found:
        return other_plt(plt_type, data, data_found)
    elif type_found:
        return wrong_plt_data(plt_type, data, data_found)
    elif data_found:
        return wrong_plt_type(plt_type, data, data_found)
    else:
        return no_plt(plt_type, data, data_found)


def get_plots(report=MAIN_REPORT):
    """
    Retrieves any plots made by the user. The general structure is as follows:

    .. code-block::python

        plots = [
        {
            'title': str,
            'xlabel': str,
            'ylabel': str,
            'legend': bool
            'data': {
                'type': str # either 'line' or 'scatter' or 'hist'
                'label': str
                # If 'hist' type
                'values': list[float]
                # If 'scatter' or 'line' type
                'x': list[float],
                'y': list[float]
            }
        }
        # ...
        ]
    """
    return get_sandbox(report=report).modules.plotting.plots




"""
Helper functions and basic setup routines for the entire module.
"""

from pedal.core.report import MAIN_REPORT
from pedal.core.commands import set_success
from pedal.sandbox.exceptions import SandboxStudentCodeException
from pedal.utilities.sorting import topological_sort


def resolve_all(set_successful=False, no_phases_is_success=False, report=MAIN_REPORT):
    """

    Args:
        set_success:
        report:
    """
    _setup_assertions(report)
    orderings = report['assertions']['relationships']
    phase_functions = report['assertions']['phase_functions']
    phase_names = report['assertions']['phases']
    phase_names = topological_sort(phase_names, orderings)
    #pprint(orderings)
    phase_success = no_phases_is_success
    for phase_name in phase_names:
        phase_success = True
        for function in phase_functions[phase_name]:
            try:
                phase_success = phase_success and (function() is not False)
            except AssertionBreak:
                phase_success = False
            except SandboxStudentCodeException:
                phase_success = False
        if not phase_success:
            break
        
    #for f in report.feedback:
    #    print("\t", f, f.mistake, f.misconception)
    if not report['assertions']['failures'] and phase_success and set_successful:
        set_success()
    
    _reset_phases(report)


def _add_phase(phase_name, function, report=MAIN_REPORT):
    phase_functions = report['assertions']['phase_functions']
    phases = report['assertions']['phases']
    if phase_name not in phase_functions:
        phase_functions[phase_name] = []
        phases.append(phase_name)
    phase_functions[phase_name].append(function)


def _add_relationships(befores, afters, report=MAIN_REPORT):
    relationships = report['assertions']['relationships']
    if None in (befores, afters):
        return
    if not isinstance(befores, (list, tuple)):
        befores = [befores]
    if not isinstance(afters, (list, tuple)):
        afters = [afters]
    for before in befores:
        if not isinstance(before, str):
            before = before.__name__
        if before not in relationships:
            relationships[before] = []
        for after in afters:
            if not isinstance(after, str):
                after = after.__name__
            relationships[before].append(after)
            
            
def _reset_phases(report=None):
    if report is None:
        report = MAIN_REPORT
    report['assertions']['relationships'].clear()
    report['assertions']['phases'].clear()
    report['assertions']['phase_functions'].clear()
    report['assertions']['failures'] = 0


def _setup_assertions(report):
    if 'assertions' not in report:
        report['assertions'] = {
            'phases': [],
            'phase_functions': {},
            'relationships': {},
            'exceptions': False,
            'failures': 0,
            'collected': [],
            # Should we batch up multiple assertion failures?
            #   The grouping mechanism is try_all
            'tabular_output': False,
        }
        report.add_hook('source.next_section.before', resolve_all)
        report.add_hook('pedal.resolvers.resolve', resolve_all)



"""
Various helper functions related to dictionaries.
"""


def extend_dictionary(d1, d2):
    """
    Helper function to create a new dictionary with the contents of the two
    given dictionaries. Does not modify either dictionary, and the values are
    copied shallowly. If there are repeats, the second dictionary wins ties.

    The function is written to ensure Skulpt compatibility.

    Args:
        d1 (dict): The first dictionary
        d2 (dict): The second dictionary
    Returns:
        dict: The new dictionary
    """
    d3 = {}
    for key, value in d1.items():
        d3[key] = value
    for key, value in d2.items():
        d3[key] = value
    return d3



"""

######### Version 1

# Ends up executing all the questions' bodies
with Pool("P1") as pool:
    with Question("QA", "Create a for loop"):
        ensure_ast("For")
    with Question("QB", "Create an if statement"):
        ensure_ast("If")
pool.choose().ask()

######### Version 2

pool = Pool("P1", [question_A, question_B])

@pool.Question("Q", "Create a for loop")
def question_A():
    ensure_ast("For")

@pool.Question("QB", "Create an if statement")
def question_B():
    ensure_ast("If")

pool.choose().ask()

######### Version 3

with Pool("P1") as pool:
    if pool.asked_question("QA", "Create a for loop"):
        ensure_ast("For")
    if pool.asked_question("QB", "Create an if statement"):
        ensure_ast("If")

"""


from pedal.core.feedback import Feedback
from pedal.core.feedback_category import FeedbackCategory, FeedbackKind
from pedal.core.report import MAIN_REPORT
from pedal.questions.graders import QuestionGrader
from pedal.questions.constants import TOOL_NAME


class Question:
    """

    """
    def __init__(self, name, instructions, tests, seed=None, report=MAIN_REPORT):
        self.name = name
        self.instructions = instructions
        self.tests = tests
        self.seed = seed
        self.report = report
        self.answered = False

    def answer(self):
        """

        """
        self.answered = True

    def ask(self):
        """

        """
        if isinstance(self.tests, QuestionGrader):
            self.tests._test(self)
        else:
            for test in self.tests:
                test(self)
        if not self.answered:
            show_question(self.instructions, report=self.report)


class show_question(Feedback):
    """

    Args:
        instructions:
        report:
    """
    category = FeedbackCategory.INSTRUCTIONS
    kind = FeedbackKind.INSTRUCTIONAL
    justification = 'Question was asked but not answered.'
    title = 'Show Question'
    priority = 'instructions'
    tool = TOOL_NAME
    valence = 0

    def __init__(self, instructions, **kwargs):
        super().__init__(message=instructions, **kwargs)



"""
A version of the resolver that still chooses one piece of feedback, but
does so per section.
"""

import sys

from pedal.core.feedback import Feedback
from pedal.core.final_feedback import parse_feedback, FinalFeedback
from pedal.resolvers import simple
from pedal.core.report import MAIN_REPORT
from pedal.resolvers.core import make_resolver
from pedal.resolvers.simple import by_priority


@make_resolver
def resolve(report=MAIN_REPORT, priority_key=by_priority):
    """
    Args:
        priority_key: The key function to sort feedbacks by
        report (Report): The report object to resolve down. Defaults to the
                         global MAIN_REPORT

    Returns
        str: A string of HTML feedback to be delivered
    """
    # Prepare feedbacks
    all_feedbacks = report.feedback
    # Group the feedback
    feedback_by_group = {}
    for feedback in all_feedbacks:
        if feedback.parent in feedback_by_group:
            feedback_by_group[feedback.parent].append(feedback)
        else:
            feedback_by_group[feedback.parent] = [feedback]
    # Process each subgroup
    finals = {}
    for group, feedbacks in feedback_by_group.items():
        feedbacks.sort(key=priority_key)
        # Create the initial final feedback
        final = FinalFeedback(success=True, score=0,
                              title=None, message=None,
                              category=Feedback.CATEGORIES.COMPLETE, label='set_success_no_errors',
                              data=[], hide_correctness=False,
                              suppressions=report.suppressions,
                              suppressed_labels=report.suppressed_labels)
        # Process each feedback in turn
        for feedback in feedbacks:
            final.merge(feedback)
        # Override empty message
        final.finalize()
        finals[group] = final
    report.result = finals
    report.resolves.append(finals)
    return finals



"""
instructions: blah blah blah

settings:
    tifa:
        enabled: True
    unit test by function (bool): Whether to test each function entirely before moving onto the
            next one, or to first check that all functions have been defined, and then
            checking their parameters, etc. Defaults to True.
    show case details (bool): Whether to show the specific args/inputs that caused a test case
            to fail.
rubric:
    functions:
        total: 100
        definition: 10
        signature: 10
        cases: 80
global:
    variables:
        name:
        type:
        value:
    inputs:
    prints:
# Sandbox, type checking
functions:
    documentation: "any" or "google"
    coverage: 100%
    tests: int
    name: do_complicated_stuff
    arity: int
    signature: int, int -> float
    signature: int, int, list[int], (int->str), dict[str:list[int]] -> list[int]
    parameters:
        name: banana
            exactly:
            regex:
            includes:
            within:
        type: int
    cases:
      - arguments (list): 5, 4
        inputs (list):
        returns (Any):
            equals: 27.3
            is:
            is not: _1
        name (str): Meaningful name for tracking purposes? Or possibly separate into label/id/code
        hint (str): Message to display to user
        prints:
            exactly:
            regex:
            startswith:
            endswith:
        plots:
# Cait
syntax:
    prevent:
        ___ + ___
# Override any of our default feedback messages
messages:
    FUNCTION_NOT_DEFINED: "Oops you missed a function"
"""
from pedal.core.commands import set_success, give_partial
from pedal.core.feedback_category import FeedbackCategory
from pedal.questions.constants import TOOL_NAME

from pedal.sandbox.commands import get_sandbox
from pedal.utilities.comparisons import equality_test

SETTING_SHOW_CASE_DETAILS = "show case details"
DEFAULT_SETTINGS = {
    SETTING_SHOW_CASE_DETAILS: True
}

EXAMPLE_DATA = {
    'functions': [{
        'name': 'do_complicated_stuff',
        'signature': 'int, int, [int] -> list[int]',
        'cases': [
            {'arguments': "5, 4, 3", 'returns': "12"},
        ]
    }]
}


class FeedbackException(Exception):
    """

    """
    def __init__(self, category, label, **fields):
        self.category = category
        self.label = label
        self.fields = fields

    def as_message(self):
        """

        Returns:

        """
        return FEEDBACK_MESSAGES[self.category][self.label].format(**self.fields)


def check_function_defined(function, function_definitions, settings=None):
    """

    Args:
        function:
        function_definitions:
        settings:

    Returns:

    """
    # 1. Is the function defined syntactically?
    # 1.1. With the right name?
    function_name = function['name']
    if function_name not in function_definitions:
        raise FeedbackException(FeedbackCategory.SPECIFICATION, 'missing_function', function_name=function_name)
    definition = function_definitions[function_name]
    return definition


def check_function_signature(function, definition, settings=None):
    """

    Args:
        function:
        definition:
        settings:

    Returns:

    """
    function_name = function['name']
    # 1.2. With the right parameters and return type?
    # 1.2.1 'arity' style - simply checks number of parameters
    if 'arity' in function or 'parameters' in function:
        expected_arity = function['arity'] if 'arity' in function else len(function['parameters'])
        actual_arity = len(definition.args.args)
        if actual_arity < expected_arity:
            raise FeedbackException(FeedbackCategory.SPECIFICATION, 'insufficient_args',
                                    function_name=function_name, expected_arity=expected_arity,
                                    actual_arity=actual_arity)
        elif actual_arity > expected_arity:
            raise FeedbackException(FeedbackCategory.SPECIFICATION, 'excessive_args',
                                    function_name=function_name, expected_arity=expected_arity,
                                    actual_arity=actual_arity)
    # 1.2.2 'parameters' style - checks each parameter's name and type
    if 'parameters' in function:
        expected_parameters = function['parameters']
        actual_parameters = definition.args.args
        for expected_parameter, actual_parameter in zip(expected_parameters, actual_parameters):
            actual_parameter_name = get_arg_name(actual_parameter)
            if 'name' in expected_parameter:
                if actual_parameter_name != expected_parameter['name']:
                    raise FeedbackException(FeedbackCategory.SPECIFICATION, 'wrong_parameter_name',
                                            function_name=function_name,
                                            expected_parameter_name=expected_parameter['name'],
                                            actual_parameter_name=actual_parameter_name
                                            )
            if 'type' in expected_parameter:
                actual_parameter_type = parse_type(actual_parameter)
                # TODO: Handle non-string expected_parameter types (dict)
                expected_parameter_type = parse_type_value(expected_parameter['type'], True)
                if not type_check(expected_parameter_type, actual_parameter_type):
                    raise FeedbackException(FeedbackCategory.SPECIFICATION, 'wrong_parameter_type',
                                            function_name=function_name,
                                            parameter_name=actual_parameter_name,
                                            expected_parameter_type=expected_parameter_type,
                                            actual_parameter_type=actual_parameter_type)
    # 1.2.3. 'returns' style - checks the return type explicitly
    if 'returns' in function:
        expected_returns = parse_type_value(function['returns'], True)
        actual_returns = parse_type(definition.returns)
        if actual_returns != "None":
            if not type_check(expected_returns, actual_returns):
                raise FeedbackException(FeedbackCategory.SPECIFICATION, "wrong_returns",
                                        function_name=function_name, expected_returns=expected_returns,
                                        actual_returns=actual_returns)
        elif expected_returns != "None":
            raise FeedbackException(FeedbackCategory.SPECIFICATION, "missing_returns",
                                    function_name=function_name, expected_returns=expected_returns)
    # 1.2.4. 'signature' style - shortcut for specifying the types
    if 'signature' in function:
        expected_signature = function['signature']
        actual_returns = parse_type(definition.returns)
        actual_parameters = ", ".join(parse_type(actual_parameter.annotation)
                                      for actual_parameter in definition.args.args)
        actual_signature = "{} -> {}".format(actual_parameters, actual_returns)
        if not type_check(expected_signature, actual_signature):
            raise FeedbackException(FeedbackCategory.SPECIFICATION, "wrong_signature",
                                    function_name=function_name, expected_signature=expected_signature,
                                    actual_signature=actual_signature)
    # All good here!
    return True


def check_function_value(function, values, settings):
    """
    2. Does the function exist in the data?

    :param function:
    :param values:
    :param settings:
    :return:
    """
    function_name = function['name']
    # 2.1. Does the name exist in the values?
    if function_name not in values:
        raise FeedbackException(FeedbackCategory.SPECIFICATION, "function_not_available", function_name=function_name)
    function_value = values[function_name]
    # 2.2. Is the name bound to a callable?
    if not callable(function_value):
        raise FeedbackException(FeedbackCategory.SPECIFICATION, "name_is_not_function", function_name=function_name)
    # All good here
    return function_value


class TestCase:
    """

    """
    CASE_COUNT = 0
    def __init__(self, function_name, case_name):
        self.function_name = function_name
        if case_name is None:
            self.case_name = str(TestCase.CASE_COUNT)
            TestCase.CASE_COUNT += 1
        else:
            self.case_name = case_name
        self.arguments, self.has_arguments = [], False
        self.inputs, self.has_inputs = [], False
        self.error, self.has_error = None, False
        self.message, self.has_message = None, False
        self.expected_prints, self.has_expected_prints = None, False
        self.expected_returns, self.has_expected_returns = None, False
        self.prints = []
        self.returns = None
        self.success = True

    def add_message(self, message):
        """

        Args:
            message:
        """
        self.message = message
        self.has_message = True

    def add_inputs(self, inputs):
        """

        Args:
            inputs:
        """
        if not isinstance(inputs, list):
            inputs = [inputs]
        self.inputs = inputs
        self.has_inputs = True

    def add_arguments(self, arguments):
        """

        Args:
            arguments:
        """
        if not isinstance(arguments, list):
            arguments = [arguments]
        self.arguments = arguments
        self.has_arguments = True

    def add_error(self, error):
        """

        Args:
            error:
        """
        self.error = error
        self.has_error = True
        self.success = False

    def add_expected_prints(self, prints):
        """

        Args:
            prints:
        """
        self.expected_prints = prints
        self.has_expected_prints = True

    def add_expected_returns(self, returns):
        """

        Args:
            returns:
        """
        self.expected_returns = returns
        self.has_expected_returns = True

    def add_prints_returns(self, prints, returns):
        """

        Args:
            prints:
            returns:
        """
        self.prints = prints
        self.returns = returns

    def fail(self):
        """

        """
        self.success = False

def check_case(function, case, student_function):
    """

    :param function:
    :param case:
    :param student_function:
    :return: status, arg, input, error, output, return, message
    """
    function_name = function['name']
    test_case = TestCase(function_name, case.get('name'))
    # Get callable
    sandbox = get_sandbox(MAIN_REPORT)
    sandbox.clear_output()
    # Potential bonus message
    if 'message' in case:
        test_case.add_message(case['message'])
    # Queue up the the inputs
    if 'inputs' in case:
        test_case.add_inputs(case['inputs'])
        sandbox.set_input(test_case.inputs)
    else:
        sandbox.clear_input()
    # Pass in the arguments and call the function
    if 'arguments' in case:
        test_case.add_arguments(case['arguments'])
    result = sandbox.call(function_name, *test_case.arguments)
    # Store actual values
    test_case.add_prints_returns(sandbox.output, result)
    # Check for errors
    if sandbox.exception:
        test_case.add_error(sandbox.exception)
    # 4. Check out the output
    if 'prints' in case:
        test_case.add_expected_prints(case['prints'])
        if not output_test(sandbox.output, case['prints'], False, .0001):
            test_case.fail()
    # 5. Check the return value
    if 'returns' in case:
        test_case.add_expected_returns(case['returns'])
        if not equality_test(result, case['returns'], True, .0001):
            test_case.fail()
    # TODO: Check the plots
    # Return results
    return test_case


# TODO: blockpy-feedback-unit => pedal-test-cases in BlockPy Client
TEST_TABLE_TEMPLATE = """<table class='pedal-test-cases table table-sm table-bordered table-hover'>
    <tr class='table-active'>
        <th></th>
        <th>Arguments</th>
        <th>Expected</th>
        <th>Returned</th>
    </tr>
    {body}
</table>"""
TEST_TABLE_FOOTER = "</table>"
TEST_TABLE_ROW_HEADER = "<tr class='table-active'>"
TEST_TABLE_ROW_NORMAL = "<tr>"
TEST_TABLE_ROW_FOOTER = "</tr>"
TEST_TABLE_ROW_INFO = "<tr class='table-info'>"
GREEN_CHECK = "        <td class='green-check-mark'>&#10004;</td>"
RED_X = "        <td>&#10060;</td>"
CODE_CELL = "        <td><code>{}</code></td>"
COLUMN_TITLES = ["", "Arguments", "Inputs", "Errors", "Expected", "Expected", "Returned", "Printed"]


def make_table(cases):
    """

    Args:
        cases:

    Returns:

    """
    body = []
    for case in cases:
        body.append("    <tr>")
        body.append(GREEN_CHECK if case.success else RED_X)
        body.append(CODE_CELL.format(", ".join(repr(arg) for arg in case.arguments)))
        if case.has_error:
            body.append("        <td colspan='2'>Error: <code>{}</code></td>".format(str(case.error)))
        else:
            body.append(CODE_CELL.format(repr(case.expected_returns)))
            body.append(CODE_CELL.format(repr(case.returns)))
        if not case.success and case.has_message:
            body.append("    </tr><tr><td colspan='4'>{}</td>".format(case.message))
        body.append("    </tr>")
    body = "\n".join(body)
    return TEST_TABLE_TEMPLATE.format(body=body)
    #if ((any(args) and any(inputs)) or
    #        (any(expected_outputs) and any(expected_returns)) or
    #        (any(actual_outputs) and any(actual_returns))):
    #    # Complex cells
    #    pass
    #else:
    # Simple table
    # Make header

    # row_mask = [True, any(args), any(inputs), False,
    #             any("returns" in reason for reason in reasons),
    #             any("prints" in reason for reason in reasons),
    #             any("returns" in reason for reason in reasons),
    #             any("prints" in reason for reason in reasons)]
    # header_cells = "".join("<th>{}</th>".format(title) for use, title in zip(row_mask, COLUMN_TITLES) if use)
    # body = [TEST_TABLE_ROW_HEADER.format(header_cells)]
    # for case in zip(
    #         statuses, args, inputs, errors, actual_outputs, actual_returns,
    #         expected_outputs, expected_returns):
    #     status, case = case[0], case[1:]
    #     print(row_mask[1:], case)
    #     def make_code(values):
    #         if values == None:
    #             return "<code>None</code>"
    #         elif isinstance(values, int):
    #             return "<code>{!r}</code>".format(values)
    #         else:
    #             return ", ".join("<code>{}</code>".format(repr(value)) for value in values)
    #     body.append(
    #         TEST_TABLE_ROW_NORMAL+
    #         (GREEN_CHECK if case[0] else RED_X)+
    #         "\n".join("    <td>{}</td>".format(make_code(values))
    #                   for use, values in zip(row_mask[1:], case) if use)+
    #         "</tr>\n"
    #     )
    # # Make each row
    # table = "{}\n{}\n{}".format(TEST_TABLE_HEADER, "\n    ".join(body), TEST_TABLE_FOOTER)
    # return table


def check_cases(function, student_function, settings):
    """

    Args:
        function:
        student_function:
        settings:
    """
    function_name = function['name']
    if 'cases' in function:
        cases = function['cases']
        test_cases = [check_case(function, case, student_function) for case in cases]
        success_cases = sum(test.success for test in test_cases)
        if success_cases < len(cases):
            if settings[SETTING_SHOW_CASE_DETAILS]:
                table = make_table(test_cases)
                raise FeedbackException(FeedbackCategory.SPECIFICATION, "failed_test_cases",
                                        function_name=function_name,
                                        cases_count=len(cases), failure_count=len(cases)-success_cases,
                                        table=table)
            else:
                raise FeedbackException(FeedbackCategory.SPECIFICATION, "failed_test_cases_count",
                                        function_name=function_name,
                                        cases_count=len(cases), failure_count=len(cases) - success_cases)


def get_arg_name(node):
    """

    Args:
        node:

    Returns:

    """
    name = node.id
    if name is None:
        return node.arg
    else:
        return name


def load_question(data):
    """

    :param data:
    :return:
    """
    ast = parse_program()
    student_data = commands.get_student_data()
    # Check that there aren't any invalid syntactical structures
    # Get all of the function ASTs in a dictionary
    function_definitions = {definition._name: definition
                            for definition in ast.find_all("FunctionDef")}
    settings = DEFAULT_SETTINGS.copy()
    settings.update(data.get('settings', {}))
    rubric = settings.get('rubric', {})
    function_points = 0
    if 'functions' in data:
        function_rubric = rubric.get('functions', {})
        successes = []
        for function in data['functions']:
            success = False
            try:
                definition = check_function_defined(function, function_definitions, settings)
                function_points += function_rubric.get('definition', 10)
                check_function_signature(function, definition, settings)
                function_points += function_rubric.get('signature', 10)
                student_function = check_function_value(function, student_data, settings)
                function_points += function_rubric.get('value', 0)
            except FeedbackException as fe:
                yield fe.as_message(), fe.label
            else:
                try:
                    check_cases(function, student_function, settings)
                except FeedbackException as fe:
                    success_ratio = (1.0 - fe.fields['failure_count'] / fe.fields['cases_count'])
                    function_points += function_rubric.get('cases', 80*success_ratio)
                    yield fe.as_message(), fe.label
                else:
                    function_points += function_rubric.get('cases', 80)
                    success = True
            successes.append(success)
        function_points /= len(data['functions'])
        if all(successes):
            set_success()
        else:
            give_partial(function_points, tool=TOOL_NAME,
                         justification="Passed some but not all unit tests")


def check_question(data):
    """

    Args:
        data:
    """
    results = list(load_question(data))
    if results:
        message, label = results[0]
        gently(message, label=label)


def check_pool(questions):
    """

    Args:
        questions:
    """
    pass


def load_file(filename):
    """

    Args:
        filename:
    """
    pass


FEEDBACK_MESSAGES = {
    FeedbackCategory.SPECIFICATION: {
        "missing_function": "No function named `{function_name}` was found.",
        "insufficient_args": ("The function named `{function_name}` "
                              "has fewer parameters ({actual_arity}) "
                              "than expected ({expected_arity})."),
        "excessive_args": ("The function named `{function_name}` "
                           "has more parameters ({actual_arity}) "
                           "than expected ({expected_arity})."),
        # TODO: missing_parameter that checks if parameter name exists, but is in the wrong place
        "wrong_parameter_name": ("Error in definition of `{function_name}`. "
                                 "Expected a parameter named `{expected_parameter_name}`, "
                                 "instead found `{actual_parameter_name}`."),
        "wrong_parameter_type": ("Error in definition of function `{function_name}` "
                                 "parameter `{parameter_name}`. Expected `{expected_parameter_type}`, "
                                 "instead found `{actual_parameter_type}`."),
        "missing_returns": ("Error in definition of function `{function_name}` return type. "
                            "Expected `{expected_returns}`, but there was no return type specified."),
        "wrong_returns": ("Error in definition of function `{function_name}` return type. "
                          "Expected `{expected_returns}`, instead found `{actual_returns}`."),
        "wrong_signature": ("Error in definition of function `{function_name}` signature. "
                            "Expected `{expected_signature}`, instead found `{actual_signature}`."),
        "name_is_not_function": "You defined `{function_name}`, but did not define it as a function.",
        "function_not_available": ("You defined `{function_name}` somewhere in your code, "
                                   "but it was not available in the top-level scope to be called. "
                                   "Perhaps you defined it inside another function or scope?"),
        "failed_test_cases": ("I ran your function <code>{function_name}</code> on my own test cases. "
                              "It failed {failure_count}/{cases_count} of my tests.\n{table}"),
        "failed_test_cases_count": ("I ran your function <code>{function_name}</code> on my own test cases. "
                                    "It failed {failure_count}/{cases_count} of my tests."),
    }
}



"""
Commands related to navigating TIFA data.
"""
from pedal.core.report import MAIN_REPORT
from pedal.tifa.constants import TOOL_NAME as TIFA_TOOL_NAME


def tifa_analysis(code=None, report=MAIN_REPORT):
    """
    Perform the TIFA analysis and attach the results to the Report.

    Args:
        code (str or None): The code to evaluate with TIFA. If ``code`` is not
            given, then it will default to the student's main file.
        report (:class:`pedal.core.report.Report`): The Report object to
            attach results to.
    Returns:
        :py:class:`pedal.tifa.tifa_core.TifaAnalysis`: A TifaAnalysis data
            bundle containing all the information that TIFA learned.
    """
    if code is None:
        code = report.submission.main_code
    if code in report[TIFA_TOOL_NAME]['analyses']:
        return report[TIFA_TOOL_NAME]['analyses'][code]
    result = report[TIFA_TOOL_NAME]['instance'].process_code(code)
    report[TIFA_TOOL_NAME]['analyses'][code] = result
    report[TIFA_TOOL_NAME]['latest'] = result
    return result


def get_issues(category, report=MAIN_REPORT):
    """

    Args:
        category (str or Feedback): The category of Issues to retrieve.

    Returns:
        list[Feedback]: The feedback functions triggered for this issue.
    """
    if not isinstance(category, str):
        category = category.__name__
    if not report[TIFA_TOOL_NAME]['latest']:
        tifa_analysis(report=report)
    latest = report[TIFA_TOOL_NAME]['latest']
    return latest.issues.get(category, [])



import sys
import os

from unittest.mock import patch

try:
    import coverage
except ImportError:
    coverage = None

try:
    from bdb import Bdb, BdbQuit
except Exception:
    class Bdb:
        pass


    class BdbQuit:
        pass

_stdout = sys.stdout


class SandboxBasicTracer:
    """

    """
    def __init__(self):
        super().__init__()
        self.filename = "student.py"
        self.code = None

    def as_filename(self, filename, code):
        if os.path.isabs(filename):
            self.filename = filename
        else:
            self.filename = os.path.abspath(filename)
        self.code = code
        return self

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, traceback):
        pass


class SandboxCoverageTracer(SandboxBasicTracer):
    """

    """
    def __init__(self):
        super().__init__()
        if coverage is None:
            raise ImportError("The coverage package is not available.")
        self.n_missing = None
        self.n_statements = None
        self.pc_covered = None
        self.missing = set()
        self.lines = set()

    def __enter__(self):
        # Force coverage to accept the code
        self.original = coverage.python.get_python_source

        def _get_source_correctly(reading_filename):
            if reading_filename == self.filename:
                return self.code
            else:
                return self.original(reading_filename)

        self.p = patch('coverage.python.get_python_source', _get_source_correctly)
        self.p.start()
        #coverage.python.get_python_source = _get_source_correctly
        self.coverage = coverage.Coverage()
        self.coverage.start()

    def __exit__(self, exc_type, exc_val, traceback):
        self.coverage.stop()
        self.coverage.save()
        # Restore the get_python_source reader
        #coverage.python.get_python_source = self.original
        # Actually analyze the data, attach some data
        analysis = self.coverage._analyze(self.filename)
        #print(vars(self.coverage._analyze(self.filename)), file=_stdout)
        self.n_missing = analysis.numbers.n_missing
        self.n_statements = analysis.numbers.n_statements
        self.pc_covered = analysis.numbers.pc_covered
        self.missing = analysis.missing
        self.lines = analysis.statements - analysis.missing

        self.p.stop()
        self.original = None

    
    @property
    def percent_covered(self):
        """

        Returns:

        """
        return self.pc_covered


class SandboxCallTracer(SandboxBasicTracer, Bdb):
    """

    """
    def __init__(self):
        super().__init__()
        self.calls = {}

    def user_call(self, frame, argument_list):
        """

        Args:
            frame:
            argument_list:
        """
        code = frame.f_code
        name = code.co_name
        if name not in self.calls:
            self.calls[name] = []
        self.calls[name].append(code)

    def __enter__(self):
        self.reset()
        self._old_trace = sys.gettrace()
        sys.settrace(self.trace_dispatch)

    def __exit__(self, exc_type, exc_val, traceback):
        sys.settrace(self._old_trace)
        self.quitting = True
        # Return true to suppress exception (if it is a BdbQuit)
        return isinstance(exc_type, BdbQuit)


TRACER_STYLES = {
    'coverage': SandboxCoverageTracer,
    'calls': SandboxCallTracer,
    'none': SandboxBasicTracer,
}



from pedal.resolvers.core import make_resolver
from pedal.core.report import MAIN_REPORT
from pedal.core.feedback import Feedback

DEFAULT_CATEGORY_PRIORITY = [
    Feedback.CATEGORIES.INSTRUCTIONS,
    Feedback.CATEGORIES.SYNTAX,
    Feedback.CATEGORIES.MISTAKES,
    Feedback.CATEGORIES.INSTRUCTOR,
    Feedback.CATEGORIES.ALGORITHMIC,
    Feedback.CATEGORIES.RUNTIME,
    Feedback.CATEGORIES.STUDENT,
    Feedback.CATEGORIES.POSITIVE,
    Feedback.CATEGORIES.UNKNOWN
]


def resolve_feedback(feedback):
    """

    Args:
        feedback:

    Returns:

    """
    return "{title}\n{body}".format(title=feedback.title, body=feedback.message)


@make_resolver
def resolve(report=MAIN_REPORT):
    """

    Args:
        report:

    Returns:

    """
    grouped = {}
    for feedback in report.feedback:
        if feedback.category not in grouped:
            grouped[feedback.category] = []
        grouped[feedback.category].append(feedback)
    result = []
    for group, feedbacks in grouped.items():
        result.append(group.title())
        result.extend([resolve_feedback(feedback) for feedback in feedbacks])
    return "\n".join(result)



"""
A collection of types.

TODO: Make the types all respect the type heirarchy properly.
For example, SetType should not descend from ListType, that's just lazy.
Instead, create a hierachy based on form and not just function.
"""

import ast

from pedal.utilities.dictionaries import extend_dictionary


def are_literals_equal(first, second):
    """
    Determines if two literals have the exact same value.
    Checks nested structures.

    Args:
        first (Any):
        second (Any):

    Returns:
        bool: Whether or not they are equal.
    """
    if first is None or second is None:
        return False
    elif not isinstance(first, type(second)):
        return False
    else:
        if isinstance(first, LiteralTuple):
            if len(first.value) != len(second.value):
                return False
            for l, s in zip(first.value, second.value):
                if not are_literals_equal(l, s):
                    return False
            return True
        elif not isinstance(first, LiteralValue):
            return True
        else:
            return first.value == second.value


class LiteralValue:
    """
    A special literal representation of a value, used to represent access on
    certain container types.
    """

    def __init__(self, value):
        self.value = value


class LiteralNum(LiteralValue):
    """
    Used to capture indexes of containers.
    """

    def type(self):
        """

        Returns:

        """
        return NumType()


class LiteralBool(LiteralValue):
    def type(self):
        """

        Returns:

        """
        return BoolType()


class LiteralStr(LiteralValue):
    def type(self):
        """

        Returns:

        """
        return StrType()


class LiteralTuple(LiteralValue):
    def type(self):
        """

        Returns:

        """
        return TupleType(self.value)


class LiteralNone(LiteralValue):
    def type(self):
        """

        Returns:

        """
        return NoneType()


class Type:
    """
    Parent class for all other types, used to provide a common interface.

    TODO: Handle more complicated object-oriented types and custom types
        (classes).
    """
    fields = {}
    immutable = False
    singular_name = 'a type'

    def clone(self):
        """

        Returns:

        """
        return self.__class__()

    def __str__(self):
        return str(self.__class__.__name__)

    def precise_description(self):
        """

        Returns:

        """
        return self.singular_name

    def clone_mutably(self):
        """

        Returns:

        """
        if self.immutable:
            return self.clone()
        else:
            return self

    def index(self, i):
        """

        Args:
            i:

        Returns:

        """
        return self.clone()

    def iterate(self, i):
        """

        Args:
            i:

        Returns:

        """
        return self.index(i)

    def load_attr(self, attr):
        """

        Args:
            attr:
            tifa:
            callee:
            callee_position:

        Returns:

        """
        if attr in self.fields:
            return self.fields[attr]
        return UnknownType()

    def is_empty(self):
        """

        Returns:

        """
        return True

    def is_equal(self, other):
        """

        Args:
            other:

        Returns:

        """
        # TODO: Make this more sophisticated!
        if type(self) not in TYPE_LOOKUPS:
            return False
        return other in TYPE_LOOKUPS[type(self)]

    def is_instance(self, other):
        """

        Args:
            other:

        Returns:

        """
        # TODO: Implement this correctly
        return self.is_equal(other)


class UnknownType(Type):
    """
    A special type used to indicate an unknowable type.
    """
    singular_name = "an unknown type"


class RecursedType(Type):
    """
    A special type used as a placeholder for the result of a
    recursive call that we have already process. This type will
    be dominated by any actual types, but will not cause an issue.
    """


class FunctionType(Type):
    """
    
    Special values for `returns`:
        identity: Returns the first argument's type
        element: Returns the first argument's first element's type
        void: Returns the NoneType
    """
    singular_name = 'a function'

    def __init__(self, definition=None, name="*Anonymous", returns=None):
        if returns is not None and definition is None:
            if returns == 'identity':
                # TODO: Determine if these can be extracted to be constant?
                def definition(ti, ty, na, args, ca):
                    """ Identify function definition """
                    if args:
                        return args[0].clone()
                    return UnknownType()
            elif returns == 'element':
                def definition(ti, ty, na, args, ca):
                    """ Element function definition """
                    if args:
                        return args[0].index(0)
                    return UnknownType()
            elif returns == 'void':
                def definition(ti, ty, na, args, ca):
                    """ Void function definition """
                    return NoneType()
            else:
                def definition(ti, ty, na, args, ca):
                    """ Generic function """
                    return returns.clone()
        self.definition = definition
        self.name = name
        self.returns = returns

    def clone(self):
        """ Create a deep copy of this function type. """
        returns = self.returns if self.returns is None else self.returns.clone()
        return FunctionType(definition=self.definition,
                            name=self.name, returns=returns)


class ClassType(Type):
    """

    """
    singular_name = 'a class'

    def __init__(self, name):
        self.name = name
        self.fields = {}
        self.scope_id = None

    def add_attr(self, name, type):
        """

        Args:
            name:
            type:
        """
        self.fields[name] = type

    def get_constructor(self):
        """

        Returns:

        """
        i = InstanceType(self)
        return FunctionType(name='__init__', returns=i)

    def clone(self):
        """

        Returns:

        """
        return ClassType(self.name)


class InstanceType(Type):
    """

    """

    def __init__(self, parent):
        self.parent = parent
        self.fields = parent.fields

    def __str__(self):
        return "InstanceTypeOf" + str(self.parent.name)

    def clone(self):
        """

        Returns:

        """
        return InstanceType(self.parent)

    def add_attr(self, name, type):
        """

        Args:
            name:
            type:
        """
        # TODO: What if this is a type change?
        self.parent.add_attr(name, type)


class NumType(Type):
    """ Basic Numeric Type (integers and floats) """
    singular_name = 'a number'
    immutable = True

    def index(self, i):
        """ Cannot index a numeric """
        return UnknownType()


class NoneType(Type):
    singular_name = 'a None'
    immutable = True


class BoolType(Type):
    singular_name = 'a boolean'
    immutable = True


class CustomType(Type):
    def __init__(self, name):
        self.singular_name = name


class IndexableType(Type):
    """
    An abstract base class to represent a type that can be accessed via
    an integer index.
    """

    def __init__(self, subtypes=None):
        """
        Create a new instance of this type.

        Args:
            subtypes (Iterable[Type]): An iterable of types.
        """
        if subtypes is None:
            subtypes = []
        self.subtypes = subtypes
        self.empty = bool(subtypes)

    def index(self, i):
        """
        Retrieves the type at the given index of this indexable type.

        Args:
            i: Either a LiteralNum or an actual integer.

        Returns:
            Type: Whatever type was at that index.
        """
        if isinstance(i, LiteralNum):
            return self.subtypes[i.value].clone()
        else:
            return self.subtypes[i].clone()

    def clone(self):
        """
        Makes a duplicate of this type by creating a new instance of
        the type and passing in a list of cloned values.

        Returns:
            IndexableType: A copy of this type.
        """
        return type(self)(subtypes=[t.clone() for t in self.subtypes])


class TupleType(IndexableType):
    """ Type representing a tuple """
    singular_name = 'a tuple'
    immutable = True


class ListType(Type):
    """ A ListType """
    singular_name = 'a list'

    def __init__(self, subtype=None, empty=True):
        if subtype is None:
            subtype = UnknownType()
        self.subtype = subtype
        self.empty = empty

    def precise_description(self):
        """ String with all the type information. """
        if self.empty:
            descr = "an empty list"
        else:
            descr = "a list"
        if not isinstance(self.subtype, UnknownType):
            descr += " of "+self.subtype.precise_description()
        return descr

    def index(self, i):
        """

        Args:
            i:

        Returns:

        """
        if isinstance(i, LiteralNum):
            return self.subtype.clone()
        else:
            return UnknownType()

    def clone(self):
        """

        Returns:

        """
        return ListType(self.subtype.clone(), self.empty)

    def load_attr(self, attr):
        """

        Args:
            attr:
            tifa:
            callee:
            callee_position:

        Returns:

        """
        if attr == 'append':
            def _append(tifa, function_type, callee, args, position):
                if args:
                    cloned_type = ListType(subtype=args[0].clone(),
                                           empty=False)
                    if callee:
                        tifa.append_variable(callee, cloned_type, position)
                    self.empty = False
                    self.subtype = args[0]
                return NoneType()

            return FunctionType(_append, 'append')
        # Extend, Pop
        return Type.load_attr(self, attr)

    def is_empty(self):
        """

        Returns:

        """
        return self.empty


class StrType(Type):
    """

    """
    singular_name = 'a string'

    def __init__(self, empty=False):
        self.empty = empty

    def index(self, i):
        """

        Args:
            i:

        Returns:

        """
        if isinstance(i, LiteralNum):
            return StrType()
        else:
            return UnknownType()


    def is_empty(self):
        """

        Returns:

        """
        return self.empty

    fields = extend_dictionary(Type.fields, {})
    immutable = True


StrType.fields.update({
    # Methods that return strings
    "capitalize": FunctionType(name='capitalize', returns=StrType()),
    "center": FunctionType(name='center', returns=StrType()),
    "expandtabs": FunctionType(name='expandtabs', returns=StrType()),
    "join": FunctionType(name='join', returns=StrType()),
    "ljust": FunctionType(name='ljust', returns=StrType()),
    "lower": FunctionType(name='lower', returns=StrType()),
    "lstrip": FunctionType(name='lstrip', returns=StrType()),
    "replace": FunctionType(name='replace', returns=StrType()),
    "rjust": FunctionType(name='rjust', returns=StrType()),
    "rstrip": FunctionType(name='rstrip', returns=StrType()),
    "strip": FunctionType(name='strip', returns=StrType()),
    "swapcase": FunctionType(name='swapcase', returns=StrType()),
    "title": FunctionType(name='title', returns=StrType()),
    "translate": FunctionType(name='translate', returns=StrType()),
    "upper": FunctionType(name='upper', returns=StrType()),
    "zfill": FunctionType(name='zfill', returns=StrType()),
    # Methods that return numbers
    "count": FunctionType(name='count', returns=NumType()),
    "find": FunctionType(name='find', returns=NumType()),
    "rfind": FunctionType(name='rfind', returns=NumType()),
    "index": FunctionType(name='index', returns=NumType()),
    "rindex": FunctionType(name='rindex', returns=NumType()),
    # Methods that return booleans
    "startswith": FunctionType(name='startswith', returns=BoolType()),
    "endswith": FunctionType(name='endswith', returns=BoolType()),
    "isalnum": FunctionType(name='isalnum', returns=BoolType()),
    "isalpha": FunctionType(name='isalpha', returns=BoolType()),
    "isdigit": FunctionType(name='isdigit', returns=BoolType()),
    "islower": FunctionType(name='islower', returns=BoolType()),
    "isspace": FunctionType(name='isspace', returns=BoolType()),
    "istitle": FunctionType(name='istitle', returns=BoolType()),
    "isupper": FunctionType(name='isupper', returns=BoolType()),
    # Methods that return List of Strings
    "rsplit": FunctionType(name='rsplit', returns=ListType(StrType(), empty=False)),
    "split": FunctionType(name='split', returns=ListType(StrType(), empty=False)),
    "splitlines": FunctionType(name='splitlines', returns=ListType(StrType(), empty=False))
})


class FileType(Type):
    singular_name = 'a file'

    def index(self, i):
        """

        Args:
            i:

        Returns:

        """
        return StrType()

    fields = extend_dictionary(Type.fields, {
        'close': FunctionType(name='close', returns='void'),
        'read': FunctionType(name='read', returns=StrType()),
        'readlines': FunctionType(name='readlines', returns=ListType(StrType(), False))
    })

    def is_empty(self):
        """

        Returns:

        """
        return False


class DictType(Type):
    """

    """
    singular_name = 'a dictionary'

    def precise_description(self):
        """

        Returns:

        """
        base = "a dictionary"
        if self.literals:
            base += " mapping "
            # TODO: Handle recursive precise names more correctly
            base += ", ".join("{!r} to {}".format(l.value, r.precise_description())
                              for l, r in zip(self.literals, self.values))
        elif self.keys:
            keys = self.keys[0] if isinstance(self.keys, list) else self.keys
            values = self.values[0] if isinstance(self.values, list) else self.values
            base += " mapping {}".format(keys.precise_description())
            base += " to {}".format(values.precise_description())
        return base

    def __init__(self, empty=False, literals=None, keys=None, values=None):
        self.empty = empty
        self.literals = literals
        self.values = values
        self.keys = keys

    def clone(self):
        """

        Returns:

        """
        # TODO: Pretty sure these should be recursively cloning fields
        return DictType(self.empty, self.literals, self.keys, self.values)

    def is_empty(self):
        """

        Returns:

        """
        return self.empty

    def has_literal(self, l):
        """

        Args:
            l:

        Returns:

        """
        for literal, value in zip(self.literals, self.values):
            if are_literals_equal(literal, l):
                return value
        return None

    def _access_item(self, i, get_keys=False):
        if self.empty:
            return UnknownType()
        elif self.literals is not None:
            for literal, value in zip(self.literals, self.values):
                if are_literals_equal(literal, i):
                    return value.clone()
            return UnknownType()
        elif get_keys:
            # TODO: Handle multiple keys/values
            return self.keys[0].clone()
        else:
            return self.values[0].clone()

    def index(self, i):
        """

        Args:
            i:

        Returns:

        """
        return self._access_item(i, False)

    def iterate(self, i):
        """

        Args:
            i:

        Returns:

        """
        return self._access_item(i, True)

    def update_key(self, literal_key, type):
        """

        Args:
            literal_key:
            type:
        """
        self.literals.append(literal_key)
        self.values.append(type)

    def load_attr(self, attr):
        """

        Args:
            attr:
            tifa:
            callee:
            callee_position:

        Returns:

        """
        if attr == 'items':
            def _items(tifa, function_type, callee, args, position):
                if self.literals is None:
                    return ListType(TupleType([self.keys, self.values]),
                                    empty=False)
                else:
                    return ListType(TupleType([self.literals[0].type(),
                                               self.values[0]]),
                                    empty=False)

            return FunctionType(_items, 'items')
        elif attr == 'keys':
            def _keys(tifa, function_type, callee, args, position):
                if self.literals is None:
                    return ListType(self.keys, empty=False)
                else:
                    return ListType(self.literals[0].type(), empty=False)

            return FunctionType(_keys, 'keys')
        elif attr == 'values':
            def _values(tifa, function_type, callee, args, position):
                if self.literals is None:
                    return ListType(self.values, empty=False)
                else:
                    return ListType(self.values[0], empty=False)

            return FunctionType(_values, 'values')
        return Type.load_attr(self, attr)


class ModuleType(Type):
    """

    """
    singular_name = 'a module'

    def __init__(self, name="*UnknownModule", submodules=None, fields=None):
        self.name = name
        if submodules is None:
            submodules = {}
        self.submodules = submodules
        if fields is None:
            fields = {}
        self.fields = fields

    def clone(self):
        """ Create a copy of this ModuleType """
        submodules = {n: v.clone() for n, v in self.submodules.items()}
        fields = {n: v.clone() for n, v in self.fields.items()}
        return ModuleType(name=self.name, submodules=submodules, fields=fields)


class SetType(ListType):
    singular_name = 'a set'


class FrozenSetType(SetType):
    """ Type representing a frozenset """
    singular_name = 'a frozen set'
    immutable = True


class GeneratorType(ListType):
    singular_name = 'a generator'


# TODO: Move these into curriculum-ctvt

class TimeType(Type):
    singular_name = 'a time of day'


class DayType(Type):
    singular_name = 'a day of the week'


try:
    from numbers import Number
except Exception:
    Number = int

TYPE_LOOKUPS = {
    FunctionType: ('function', FunctionType, 'FunctionType'),
    ClassType: ('class', ClassType, 'ClassType'),
    InstanceType: ('instance', InstanceType, 'InstanceType'),
    NumType: ('num', int, float, complex, NumType, Number, 'NumType'),
    BoolType: ('bool', bool, BoolType, 'BoolType'),
    NoneType: ('None', None, NoneType, 'NoneType'),
    TupleType: ('tuple', tuple, TupleType, 'TupleType'),
    ListType: ('list', list, ListType, 'ListType'),
    StrType: ('str', str, StrType, 'StrType'),
    FileType: ('file', FileType, 'FileType'),
    DictType: ('dict', dict, DictType, 'DictType'),
    SetType: ('set', set, SetType, 'SetType'),
}


TYPE_STRINGS = {
    "str": StrType, "string": StrType,
    "num": NumType, "number": NumType,
    "int": NumType, "integer": NumType, "float": NumType,
    "complex": NumType,
    "bool": BoolType, "boolean": BoolType,
    "none": NoneType, "nonetype": NoneType,
    "dict": DictType, "dictionary": DictType,
    "list": ListType,
    "tuple": TupleType,
    "set": SetType,
    "file": FileType,
    "func": FunctionType, "function": FunctionType,
    "class": ClassType,
}


class TypeSpace:
    def has(self, name):
        return False

    def get(self, name):
        return None



"""
Drop-in replacement class for unittest.TestCase.
"""
from pedal.assertions.runtime import *


class PedalTestCase:
    """
    Drop-in replacement for the existing unittest.TestCase class. Emulates
    the original interface, but also supports all the Pedal features, allowing
    an easier transition to our toolset.

    TODO: Implement the entire interface
        setUp()
        tearDown()
        setUpClass()
        tearDownClass()
        skipTest()
        subTest()
        debug()
        run()
        fail()
    """

    def assertEqual(self, left, right, msg=None, **kwargs):
        """
        Test that first and second are equal.
        If the values do not compare equal, the test will fail.
        """
        return assert_equal(left, right, explanation=msg, **kwargs)

    def assertNotEqual(self, left, right, msg=None, **kwargs):
        """
        Test that first and second are not equal.
        If the values do compare equal, the test will fail.
        """
        return assert_not_equal(left, right, explanation=msg, **kwargs)

    def assertLess(self, left, right, msg=None, **kwargs):
        """
        Test that first is less than the second.
        """
        return assert_less(left, right, explanation=msg, **kwargs)

    def assertLessEqual(self, left, right, msg=None, **kwargs):
        """
        Test that first is less than or equal to the second.
        """
        return assert_less_equal(left, right, explanation=msg, **kwargs)

    def assertGreater(self, left, right, msg=None, **kwargs):
        """
        Test that first is greater than the second.
        """
        return assert_greater(left, right, explanation=msg, **kwargs)

    def assertGreaterEqual(self, left, right, msg=None, **kwargs):
        """
        Test that first is greater than or equal to the second.
        """
        return assert_greater_equal(left, right, explanation=msg, **kwargs)

    def assertIn(self, needle, haystack, msg=None, **kwargs):
        """
        Tests that the needle is in the haystack.
        """
        return assert_in(needle, haystack, explanation=msg, **kwargs)

    def assertNotIn(self, needle, haystack, msg=None, **kwargs):
        """
        Tests that the needle is not in the haystack.
        """
        return assert_not_in(needle, haystack, explanation=msg, **kwargs)

    def assertIs(self, left, right, msg=None, **kwargs):
        """
        Test that first is second.
        """
        return assert_is(left, right, explanation=msg, **kwargs)

    def assertIsNot(self, left, right, msg=None, **kwargs):
        """
        Test that first is not second.
        """
        return assert_is_not(left, right, explanation=msg, **kwargs)

    def assertIsNone(self, expr, msg=None, **kwargs):
        """
        Test that expr is None
        """
        return assert_is_none(expr, explanation=msg, **kwargs)

    def assertIsNotNone(self, expr, msg=None, **kwargs):
        """
        Test that expr is not None
        """
        return assert_is_not_none(expr, explanation=msg, **kwargs)

    def assertTrue(self, expr, msg=None, **kwargs):
        """
        Test that expr is true
        """
        return assert_true(expr, explanation=msg, **kwargs)

    def assertFalse(self, expr, msg=None, **kwargs):
        """
        Test that expr is false
        """
        return assert_false(expr, explanation=msg, **kwargs)

    def assertLengthEqual(self, sequence, length, msg=None, **kwargs):
        """
        Test that sequence has length.
        """
        return assert_length_equal(sequence, length, explanation=msg, **kwargs)

    def assertLengthNotEqual(self, sequence, length, msg=None, **kwargs):
        """
        Test that sequence does not have length.
        """
        return assert_length_not_equal(sequence, length, explanation=msg, **kwargs)

    def assertLengthLess(self, sequence, length, msg=None, **kwargs):
        """
        Test that sequence has less than length.
        """
        return assert_length_less(sequence, length, explanation=msg, **kwargs)

    def assertLengthLessEqual(self, sequence, length, msg=None, **kwargs):
        """
        Test that sequence has less or equal to length.
        """
        return assert_length_less_equal(sequence, length, explanation=msg, **kwargs)

    def assertLengthGreater(self, sequence, length, msg=None, **kwargs):
        """
        Test that sequence has greater than length.
        """
        return assert_length_greater(sequence, length, explanation=msg, **kwargs)

    def assertLengthGreaterEqual(self, sequence, length, msg=None, **kwargs):
        """
        Test that sequence has greater or equal length.
        """
        return assert_length_greater_equal(sequence, length, explanation=msg, **kwargs)

    def assertIsInstance(self, obj, cls, msg=None, **kwargs):
        """
        Test that obj is an instance of cls.
        """
        return assert_is_instance(obj, cls, explanation=msg, **kwargs)

    def assertNotIsInstance(self, obj, cls, msg=None, **kwargs):
        """
        Test that obj is not an instance of cls.
        """
        return assert_not_is_instance(obj, cls, explanation=msg, **kwargs)

    def assertAlmostEqual(self, first, second, places=7, msg=None, delta=None, **kwargs):
        """ Test that first is approximately equal to second. """
        if places is not None:
            delta = 10**(-places)
        return assert_equal(first, second, delta=delta, explanation=msg, **kwargs)

    def assertNotAlmostEqual(self, first, second, places=7, msg=None, delta=None, **kwargs):
        """ Test that first is not approximately equal to second. """
        if places is not None:
            delta = 10**(-places)
        return assert_not_equal(first, second, delta=delta, explanation=msg, **kwargs)

    def assertRegex(self, regex, text, msg=None, **kwargs):
        """
        Test that regex matches text.
        """
        return assert_regex(regex, text, explanation=msg, **kwargs)

    def assertNotRegex(self, regex, text, msg=None, **kwargs):
        """
        Test that regex does not match text.
        """
        return assert_not_regex(regex, text, explanation=msg, **kwargs)

    assertMultiLineEqual = assert_equal
    assertSequenceEqual = assert_equal
    assertListEqual = assert_equal
    assertTupleEqual = assert_equal
    assertSetEqual = assert_equal
    assertDictEqual = assert_equal



import sys

from pedal.core.commands import contextualize_report
from pedal.sandbox import run
from pedal.core.environment import Environment
from pedal.core.submission import Submission


class SandboxEnvironment(Environment):
    """
    Sandbox Environment that simply runs the students code and spits out the
    output and any runtime errors. Does not attempt to provide feedback or run
    an instructor control script.

    """
    def __init__(self, code):
        contextualize_report(Submission({'answer.py': code}))
        self.student = run()

    def resolve(self):
        """

        Returns:

        """
        print(self.student.raw_output)
        if self.student.exception:
            print(self.student.exception_formatted, file=sys.stderr)



"""
Environments are there to simplify:

* Setting up a submission
* Running "default" tools in a preconfigured way
* Adjusting output as needed

Environments
    BlockPy
    VPL
    Jupyter
    GradeScope
    WebCAT
    ...

"""


'''
def default_pipeline(tifa=False, cait=True, sandbox=True):
    next_section()
    results = []
    if tifa:
        results.append(tifa_analysis())
    if cait:
        results.append(parse_program())
    if sandbox:
        results.append(execute())
    return tuple(results)
'''

from pedal.environments.standard import setup_pedal
